/*                                                                |- Impulse -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is the Impulse utility library
 * for Impulse. (Part of the ROWS distribution)
 *
 * The Initial Developer of the Original Code is
 * Daniel Berrange. Portions created by Daniel Berrange
 * are Copyright (C) 1999 Daniel Berrange.
 * All Rights Reserved
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * channel.c
 * ---------
 *
 * Communication channel functions
 */


#include "../../../config.h"

#include "impulse/comms/channel_int.h"
#include "impulse/util/error.h"
#include "impulse/util/memory.h"

#include <signal.h>
/*                            ======--------======                           */

static impulse_bool impulse_channel_link_control(impulse_channel channel,
						 impulse_channel_service service,
						 impulse_channel_event event);

static void send_error_event(impulse_channel channel,
			     impulse_uint sequence,
			     impulse_uint error,
			     impulse_ubyte service,
			     impulse_ubyte opcode,
			     impulse_unistring info);

static void send_auth_request(impulse_channel channel,
			      impulse_unistring protocol,
			      impulse_uint ndata,
			      impulse_ubyte *data);
static void send_auth_reply(impulse_channel channel,
			    impulse_uint sequence,
			    impulse_uint phase,
			    impulse_uint ndata,
			    impulse_ubyte *data);

static void send_activate_request(impulse_channel channel,
				  impulse_unistring name,
				  impulse_uint nver,
				  struct _impulse_channel_version *ver,
				  impulse_ubyte opcode,
				  impulse_bool auth);

static void send_activate_reply(impulse_channel channel,
				impulse_uint sequence,
				impulse_ubyte service,
				struct _impulse_channel_version *ver,
				impulse_ubyte opcode);

static void send_deactivate_request(impulse_channel channel,
				    impulse_ubyte opcode);
static void send_deactivate_reply(impulse_channel channel,
				  impulse_uint service,
				  impulse_bool accept);

static void send_query_request(impulse_channel channel,
			       impulse_unistring name);
static void send_query_reply(impulse_channel channel,
			     impulse_uint sequence,
			     impulse_uint nver,
			     struct _impulse_channel_version *ver);

static void send_ping_request(impulse_channel channel,
			      impulse_uint secs,
			      impulse_uint usecs);
static void send_ping_reply(impulse_channel,
			    impulse_uint sequence,
			    impulse_uint secs,
			    impulse_uint usecs);

static void process_error_event(impulse_channel channel,
				impulse_message message);
static void process_auth_request(impulse_channel channel,
				 impulse_message message);
static void process_auth_reply(impulse_channel channel,
			       impulse_message message);
static void process_activate_request(impulse_channel channel,
				     impulse_message message);
static void process_activate_reply(impulse_channel channel,
				   impulse_message message);
static void process_deactivate_request(impulse_channel channel,
				       impulse_message message);
static void process_deactivate_reply(impulse_channel channel,
				     impulse_message message);
static void process_query_request(impulse_channel channel,
				  impulse_message message);
static void process_query_reply(impulse_channel channel,
				impulse_message message);
static void process_ping_request(impulse_channel channel,
				 impulse_message message);
static void process_ping_reply(impulse_channel channel,
			       impulse_message message);

/*                            ======--------======                           */

static impulse_int nexteventid = 1;

static struct _impulse_unistring link_control_name = 
{
  11,11,
  (impulse_unichar[]) {
    (impulse_unichar)'L',
    (impulse_unichar)'i',
    (impulse_unichar)'n',
    (impulse_unichar)'k',
    (impulse_unichar)'C',
    (impulse_unichar)'o',
    (impulse_unichar)'n',
    (impulse_unichar)'t',
    (impulse_unichar)'r',
    (impulse_unichar)'o',
    (impulse_unichar)'l',
    (impulse_unichar)'\0'
  }
};

static struct _impulse_channel_service link_control = 
{
  &link_control_name, 
  1,
  (struct _impulse_channel_version[]) {
    (struct _impulse_channel_version){ 7,9 },
    (struct _impulse_channel_version){ 5,6 }
  },
  impulse_channel_link_control
};

typedef enum _impulse_control_opcode
{
  CONTROL_ERROR_EVENT = 0,
  CONTROL_AUTH_REQUEST,
  CONTROL_AUTH_REPLY,
  CONTROL_ACTIVATE_REQUEST,
  CONTROL_ACTIVATE_REPLY,
  CONTROL_DEACTIVATE_REQUEST,
  CONTROL_DEACTIVATE_REPLY,
  CONTROL_PING_REQUEST,
  CONTROL_PING_REPLY,
  CONTROL_QUERY_REQUEST,
  CONTROL_QUERY_REPLY,
  CONTROL_DETAILS_REQUEST,
  CONTROL_DETAILS_REPLY
} impulse_control_opcode;

typedef enum _impulse_control_error
{
  CONTROL_INVALID_SERVICE = 0,
  CONTROL_INVALID_MESSAGE,
  CONTROL_AUTH_NOT_SUPPORTED,
  CONTROL_AUTH_REQUIRED,
  CONTROL_SERVICE_NOT_SUPPORTED,
  CONTROL_VERSION_NOT_SUPPORTED,
  CONTROL_SERVICE_ACTIVE,
  CONTROL_OPCODE_ACTIVE
} impulse_control_error;


typedef void (*processor)(impulse_channel, impulse_message);

static processor control_process[] = {
  process_error_event,
  process_auth_request,
  process_auth_reply,
  process_activate_request,
  process_activate_reply,
  process_deactivate_request,
  process_deactivate_reply,
  process_query_request,
  process_query_reply,
  process_ping_request,
  process_ping_reply
};


/*                            ======--------======                           */

/*
 * Some useful functions for writing complex datatypes to messages
 */

static void version_writer(impulse_message message,
			   impulse_uint length,
			   void *arg)
{
  impulse_int i;
  struct _impulse_channel_version *version = (struct _impulse_channel_version *)arg;

  for (i = 0 ; i < length ; i++)
    impulse_message_write(message, "BB", version[i].major, version[i].minor);
}

static void version_reader(impulse_message message,
			   impulse_uint length,
			   void *arg)
{
  impulse_int i;
  struct _impulse_channel_version **version = (struct _impulse_channel_version **)arg;
  *version = impulse_alloc(sizeof(struct _impulse_channel_version)*length);

  for (i = 0 ; i < length ; i++)
    impulse_message_read(message, "BB", &version[i]->major, &version[i]->minor);
}

static struct _impulse_message_formatter version_formatter = 
{
  'v',
  version_writer,
  version_reader,
};

static void unistring_writer(impulse_message message,
			     impulse_uint length,
			     void *arg)
{
  impulse_int i;
  impulse_unistring *str = (impulse_unistring *)arg;
  for (i = 0 ; i < length ; i++)
    impulse_message_write(message, "xu", str[i]->length+1, str[i]->data);
}

static void unistring_reader(impulse_message message,
			     impulse_uint length,
			     void *arg)
{
  impulse_int i;
  impulse_unistring **unistring = (impulse_unistring **)arg;
  *unistring = impulse_alloc(sizeof(impulse_unistring)*length);

  for (i = 0 ; i < length ; i++)
  {
    (*unistring)[i] = impulse_unistring_create_space(0);
    impulse_message_read(message, "xu", &(*unistring)[i]->length, &(*unistring)[i]->data);
    (*unistring)[i]->capacity = --(*unistring)[i]->length;
  }
}

static struct _impulse_message_formatter unistring_formatter = 
{
  'r',
  unistring_writer,
  unistring_reader,
};



/*                            ======--------======                           */

/*
 * Messaege processing
 */

static impulse_bool impulse_channel_dispatch(impulse_event *ev,
					     void *data)
{
  impulse_channel channel = (impulse_channel)data;
  _p_impulse_channel internal = channel->internal;
  _p_impulse_channel_service service = NULL;
  struct _impulse_channel_event event;
  impulse_message message = (impulse_message)ev->data.user.data;
  
  printf("Dispatching event %d %d\n", message->header.service, message->header.opcode);

  if ((message->header.service >= internal->nincoming) ||
      !internal->incoming[message->header.service])
  {
    send_error_event(channel, 0, CONTROL_INVALID_SERVICE, message->header.service, 
		     message->header.opcode, NULL);
    goto cleanup;
  }

  service = internal->incoming[message->header.service];

  /* See if service is activated */
  /* Is this test needed, given that if it is not active it
   * will be caught by the prev test ? */
  if (!service->active)
  {
    send_error_event(channel, 0, CONTROL_INVALID_SERVICE, message->header.service, 
		     message->header.opcode, NULL);
    goto cleanup;
  }

  /* See is service is authenticated */
  if (!service->auth)
  {
    /* Allow first 5 messages of link control service through */
    if (service->incoming > 0 ||
	message->header.opcode > 4)
    {
      send_error_event(channel, 0, CONTROL_AUTH_REQUIRED, message->header.service,
		       message->header.opcode, NULL);
      goto cleanup;
    }
  }

  event.type = IMPULSE_CHANNEL_MESSAGE;
  event.data.message = *message;
  service->external->handler(channel, service->external, &event);

  cleanup:
  impulse_message_destroy(message);
  return 1;
}

  
static impulse_bool impulse_channel_process(impulse_transport transport,
					    impulse_transport_event event,
					    void *data)
{
  impulse_channel channel = (impulse_channel)data;
  _p_impulse_channel internal = channel->internal;
  impulse_int pending = impulse_transport_pending(channel->transport);

  if (!pending)
  {
    impulse_unistring desc = impulse_unistring_create_utf("No data pending (end of file)", 8);
    impulse_error_raise(IMPULSE_ERROR_CRITICAL, desc, NULL);
  }
  else if (pending < 0)
  {
    impulse_unistring desc = impulse_unistring_create_utf("Bad juju", 8);
    impulse_error_raise(IMPULSE_ERROR_CRITICAL, desc, NULL);
  }
  else
  {
    /* Make sure we use up all pending data, otherwise select
     * will wastefully spin */
    while (pending)
    {
      impulse_int need = internal->need;
      impulse_int got = 0;
      need = (need > pending) ? pending : need;

      if ((got = impulse_transport_receive(channel->transport, internal->offset, need, NULL)) < 0)
      {
	impulse_unistring desc = impulse_unistring_create_utf("More bad juju", 8);
	impulse_error_raise(IMPULSE_ERROR_CRITICAL, desc, NULL);
      }

      internal->offset += got;
      internal->need -= got;

      if (!internal->need) /* We're at the end of the block */
      {
	if (!internal->message->payload) /* We're finished reading a header */
	{
	  impulse_message_header header = &internal->message->header;
	  header->service = impulse_ubyte_be(header->service);
	  header->opcode = impulse_ubyte_be(header->opcode);
	  header->length = impulse_ushort_be(header->length);
	  
	  internal->message->payload = impulse_alloc(header->length);
	  internal->message->offset = internal->message->payload;
	  internal->need = header->length;
	  internal->offset = internal->message->payload;
	}
	else /* We're finished reading a payload */
	{
	  /* Put the finished message in the queue */
	  impulse_event *event = impulse_alloc(sizeof(impulse_event));
	  event->type = IMPULSE_EVENT_COMMS;
	  event->id = 0;
	  event->data.user.object = internal->eventid;
	  event->data.user.type = 0;
	  event->data.user.data = internal->message;

	  impulse_event_trigger(event);
	  
	  /* Setup the new message header */
	  internal->message = impulse_alloc(sizeof(struct _impulse_message));
	  internal->message->payload = NULL;
	  internal->offset = &internal->message->header;
	  internal->need = sizeof(struct _impulse_message_header);
	}
      }
      pending -= got;
    }
  }
  return 1;
}


static impulse_bool impulse_channel_link_control(impulse_channel channel,
						 impulse_channel_service service,
						 impulse_channel_event event)
{
  if (event->data.message.header.opcode > CONTROL_DETAILS_REPLY)
    send_error_event(channel, 0, CONTROL_INVALID_MESSAGE, 0, 
		     event->data.message.header.opcode, NULL);
  else
    control_process[event->data.message.header.opcode](channel, &event->data.message);
  return 0; 
}


/*                            ======--------======                           */

/*
 * Public interface 
 */

impulse_channel impulse_channel_create(impulse_transport transport,
				       impulse_bool initiator)
{
  impulse_channel channel = impulse_alloc(sizeof(struct _impulse_channel));
  _p_impulse_channel internal = impulse_alloc(sizeof(struct __p_impulse_channel));
  
  internal->message = impulse_alloc(sizeof(struct _impulse_message));
  internal->offset = &internal->message->header;
  internal->need = sizeof(struct _impulse_message_header);
  internal->message->payload = NULL;

  internal->eventid = nexteventid++;
  internal->services = impulse_hash_create();
  internal->nincoming = 1;
  internal->noutgoing = 1;
  internal->incoming = impulse_alloc(sizeof(_p_impulse_channel_service));
  internal->outgoing = impulse_alloc(sizeof(_p_impulse_channel_service));
  
  channel->transport = transport;
  channel->internal = internal;

  impulse_transport_handle(transport,
			   impulse_channel_process,
			   channel);
  
  impulse_event_register(IMPULSE_EVENT_COMMS,
			 internal->eventid,
			 impulse_channel_dispatch,
			 channel);

  internal->incoming[0] = impulse_alloc(sizeof(struct __p_impulse_channel_service));
  internal->outgoing[0] = internal->incoming[0];
  internal->incoming[0]->external = &link_control;
  internal->incoming[0]->incoming = 0;
  internal->incoming[0]->outgoing = 0;
  internal->incoming[0]->version.major = 0;
  internal->incoming[0]->version.minor = 0;
  
  internal->incoming[0]->init = 1;
  internal->incoming[0]->auth = 1;
  internal->incoming[0]->active = 1;

  impulse_hash_store(internal->services, link_control.name, internal->incoming[0]);

  if (initiator)
    send_activate_request(channel, link_control.name, link_control.nversion,
			  link_control.version, 0, 1);
  
  return channel;
}


void impulse_channel_destroy(impulse_channel channel)
{
  _p_impulse_channel internal = (_p_impulse_channel)channel->internal;
  
  impulse_transport_destroy(channel->transport);
  
  if (internal->message)
  {
    if (internal->message->payload)
      impulse_free(internal->message->payload);
    impulse_free(internal->message);
  }
    
  impulse_hash_destroy(internal->services);
  
  impulse_free(internal->incoming);
  impulse_free(internal->outgoing);
  
  impulse_free(internal);
  impulse_free(channel);
}


impulse_int impulse_channel_send(impulse_channel channel,
				 impulse_message message)
{
  _p_impulse_channel internal = (_p_impulse_channel)channel->internal;
  struct _impulse_message_header header;

  printf("Sending message %d %d\n", message->header.service, message->header.opcode);
  if (message->header.service >= internal->noutgoing ||
      !internal->outgoing[message->header.service])
  {
    printf("Send on unknown service\n");
    return 0;
  }
  
  header.service = impulse_ubyte_be(message->header.service);
  header.opcode = impulse_ubyte_be(message->header.opcode);
  header.length = impulse_ushort_be(message->header.length);
  
  impulse_transport_send(channel->transport, (impulse_ubyte *)&header, sizeof(header), NULL);
  impulse_transport_send(channel->transport, (impulse_ubyte *)message->payload, message->header.length, NULL);
  return 0;
}


void impulse_channel_register(impulse_channel channel,
			      impulse_channel_service service)
{
  _p_impulse_channel internal = (_p_impulse_channel)channel->internal;
  _p_impulse_channel_service entry;
  
  if (impulse_hash_recover(internal->services, service->name))
  {
    impulse_unistring desc = impulse_unistring_create_utf("Service already registered", 8);
    impulse_error_raise(IMPULSE_ERROR_CRITICAL, desc, NULL);
  }
  
  entry = impulse_alloc(sizeof(struct __p_impulse_channel_service));
  entry->external = service;
  entry->incoming = 0;
  entry->outgoing = 0;
  entry->version.major = 0;
  entry->version.minor = 0;
  
  entry->init = 0;
  entry->auth = 0;
  entry->active = 0;

  impulse_hash_store(internal->services, service->name, entry);
}


void impulse_channel_deregister(impulse_channel channel,
				impulse_channel_service service)
{
  _p_impulse_channel internal = (_p_impulse_channel)channel->internal;
  _p_impulse_channel_service entry = impulse_hash_remove(internal->services, service->name);
  
  impulse_free(entry);
}


void impulse_channel_activate(impulse_channel channel,
			      impulse_channel_service service)
{
  _p_impulse_channel internal = (_p_impulse_channel)channel->internal;
  _p_impulse_channel_service entry = impulse_hash_recover(internal->services, service->name);
  impulse_bool found;
  impulse_int i;
  
  if (entry->active)
    return;
  
  /*
   * Find a spare slot in the outgoing opcode map 
   */
  for (i = 0 ; i < internal->noutgoing ; i++)
  {
    if (!internal->outgoing[i])
    {
      entry->outgoing = i;
      found = 1;
      break;
    }
  }
  if (!found)
  {
    entry->outgoing = internal->noutgoing++;
    internal->outgoing = impulse_realloc(internal->outgoing,
					 sizeof(_p_impulse_channel_service)*(internal->noutgoing));
    internal->outgoing[entry->outgoing] = entry;
  }
  
  entry->init = 1;
  entry->auth = 0;
  entry->active = 0;
  printf("Sending activate %d %d %d\n", entry->outgoing, internal->nincoming, internal->noutgoing);
  send_activate_request(channel, service->name, service->nversion,
			service->version, entry->outgoing, 1);
}


void impulse_channel_deactivate(impulse_channel channel,
				impulse_channel_service service)
{
  _p_impulse_channel internal = (_p_impulse_channel)channel->internal;
  _p_impulse_channel_service entry = impulse_hash_recover(internal->services, service->name);
  
  if (!entry->active)
    return;

  send_deactivate_request(channel, entry->outgoing);
#if 0
  internal->outgoing[entry->outgoing] = NULL;
  internal->incoming[entry->incoming] = NULL;
  entry->active = 0;
  entry->auth = 0;
  entry->outgoing = 0;
  entry->incoming = 0;
#endif
}


/*                            ======--------======                           */

/*
 * Link control private interface 
 */

static void send_error_event(impulse_channel channel,
			     impulse_uint sequence,
			     impulse_uint error,
			     impulse_ubyte service,
			     impulse_ubyte opcode,
			     impulse_unistring info)
{
  impulse_message message;
  impulse_uint length;
  impulse_message_formatter formatter = &unistring_formatter;

  if (info)
  {
    length = 4 + 1 + 1 + 4 + 4 + (2 * (info->length+1));
    
    message = impulse_message_create(0, CONTROL_ERROR_EVENT, length);
    impulse_message_writef(message, 1, &formatter, "IIBBr", sequence, 
			   error, service, opcode, &info);
  }
  else
  {
    /* Just in case a null string is passed in */
    length = 4 + 1 + 1 + 4 + 4;
    message = impulse_message_create(0, CONTROL_ERROR_EVENT, length);
    impulse_message_write(message, "IIBBI", sequence, 
			  error, service, opcode, 0);
  }
  
  impulse_channel_send(channel, message);
  impulse_message_destroy(message);
}


static void send_auth_request(impulse_channel channel,
			      impulse_unistring protocol,
			      impulse_uint ndata,
			      impulse_ubyte *data)
{
  impulse_message message;
  impulse_uint length;
  impulse_message_formatter formatter = &unistring_formatter;
  
  length = 4 + (2 * (protocol->length + 1)) + 4 + ndata;
  
  message = impulse_message_create(0, CONTROL_AUTH_REQUEST, length);
  impulse_message_writef(message, 1, &formatter, "rxB", &protocol,
			 ndata, data);
  impulse_channel_send(channel, message);
  impulse_message_destroy(message);
}


static void send_auth_reply(impulse_channel channel,
			    impulse_uint sequence,
			    impulse_uint phase,
			    impulse_uint ndata,
			    impulse_ubyte *data)
{
  impulse_message message;
  impulse_uint length;
  
  length = 4 + 4 + 4 + ndata;
  
  message = impulse_message_create(0, CONTROL_AUTH_REPLY, length);
  impulse_message_write(message, "IIxB",sequence, phase, ndata, data);
  impulse_channel_send(channel, message);
  impulse_message_destroy(message);
}


static void send_activate_request(impulse_channel channel,
				  impulse_unistring name,
				  impulse_uint nver,
				  struct _impulse_channel_version *ver,
				  impulse_ubyte opcode,
				  impulse_bool auth)
{
  impulse_message message;
  impulse_uint length;
  impulse_message_formatter formatter[2] = { &version_formatter, &unistring_formatter };
  
  length = 4 + (sizeof(impulse_unichar) * (name->length + 1)) +
    4 + (2 * sizeof(impulse_ubyte) * nver) + 2;
  impulse_error_raise(IMPULSE_ERROR_WARNING, name, NULL);
  printf("Sending activate %d\n", opcode);
  message = impulse_message_create(0, CONTROL_ACTIVATE_REQUEST, length);
  impulse_message_writef(message, 2, formatter, "rxvBo",
			 &name, nver, ver,
			 opcode, auth);
  impulse_channel_send(channel, message);
  impulse_message_destroy(message);
}


static void send_activate_reply(impulse_channel channel,
				impulse_uint sequence,
				impulse_ubyte service,
				struct _impulse_channel_version *ver,
				impulse_ubyte opcode)
{
  impulse_message message;
  impulse_uint length;
  impulse_message_formatter formatter = &version_formatter;
  
  length = 4 + 1 + 2 + 1;

  message = impulse_message_create(0, CONTROL_ACTIVATE_REPLY, length);
  impulse_message_writef(message, 1, &formatter, "IBvB",sequence, service, ver, opcode);
  impulse_channel_send(channel, message);
  impulse_message_destroy(message);
}


static void send_deactivate_request(impulse_channel channel,
				    impulse_ubyte opcode)
{
  impulse_message message;
  impulse_uint length;
  
  length = 1;
  
  message = impulse_message_create(0, CONTROL_DEACTIVATE_REQUEST, length);
  impulse_message_write(message, "B", opcode);
  impulse_channel_send(channel, message);
  impulse_message_destroy(message);
}


static void send_deactivate_reply(impulse_channel channel,
				  impulse_uint service,
				  impulse_bool accept)
{
  impulse_message message;
  impulse_uint length;
  
  length = 4 + 1;
  
  message = impulse_message_create(0, CONTROL_DEACTIVATE_REPLY, length);
  impulse_message_write(message, "Io",service, accept);
  impulse_channel_send(channel, message);
  impulse_message_destroy(message);
}


static void send_query_request(impulse_channel channel,
			       impulse_unistring name)
{
  impulse_message message;
  impulse_uint length;
  impulse_message_formatter formatter = &unistring_formatter;
  
  length = 4 + (2 * (name->length + 1));
  
  message = impulse_message_create(0, CONTROL_QUERY_REQUEST, length);
  impulse_message_writef(message, 1, &formatter, "r", &name);
  impulse_channel_send(channel, message);
  impulse_message_destroy(message);
}


static void send_query_reply(impulse_channel channel,
			     impulse_uint sequence,
			     impulse_uint nver,
			     struct _impulse_channel_version *ver)
{
  impulse_message message;
  impulse_uint length;
  impulse_message_formatter formatter = &version_formatter;
  
  length = 4 + 4 + (nver * 2);
  
  message = impulse_message_create(0, CONTROL_QUERY_REPLY, length);
  impulse_message_writef(message, 1, &formatter, "IIxv", sequence, nver, ver);
  impulse_channel_send(channel, message);
  impulse_message_destroy(message);
}


static void send_ping_request(impulse_channel channel,
			      impulse_uint secs,
			      impulse_uint usecs)
{
  impulse_message message;
  impulse_uint length;
  
  length = 4 + 4;
  
  message = impulse_message_create(0, CONTROL_PING_REQUEST, length);
  impulse_message_write(message, "II", secs, usecs);
  impulse_channel_send(channel, message);
  impulse_message_destroy(message);
}


static void send_ping_reply(impulse_channel channel,
			    impulse_uint sequence,
			    impulse_uint secs,
			    impulse_uint usecs)
{
  impulse_message message;
  impulse_uint length;
  
  length = 4 + 4 + 4;
  
  message = impulse_message_create(0, CONTROL_PING_REPLY, length);
  impulse_message_write(message, "III", sequence, secs, usecs);
  impulse_channel_send(channel, message);
  impulse_message_destroy(message);
}


static void process_error_event(impulse_channel channel,
				impulse_message message)
{
  impulse_message_formatter formatter = &unistring_formatter;
  impulse_uint sequence;
  impulse_uint error;
  impulse_ubyte service;
  impulse_ubyte opcode;
  impulse_unistring *info;
  impulse_unistring desc;
  impulse_char buffer[1024];

  impulse_message_readf(message, 1, &formatter, "IIBBr",
			&sequence, &error, &service,
			&opcode, &info);
  
  sprintf(buffer, "Channel error %d, service: %d, opcode %d", error, service, opcode);
  desc = impulse_unistring_create_utf(buffer, 8);
  impulse_error_raise(IMPULSE_ERROR_WARNING, desc, info[0]);
  impulse_unistring_destroy(desc);
  impulse_free(info[0]);
  impulse_free(info);
}


static void process_auth_request(impulse_channel channel,
				 impulse_message message)
{
  impulse_message_formatter formatter = &unistring_formatter;
  impulse_unistring *protocol;
  impulse_uint ndata;
  impulse_ubyte *data;
  printf("Process auth request\n");
  impulse_message_readf(message, 1, &formatter, "rxB",
			&protocol, &ndata, &data);
  /* What the hell do we do now? */
}


static void process_auth_reply(impulse_channel channel,
			       impulse_message message)
{
  impulse_uint sequence;
  impulse_uint phase;
  impulse_uint ndata;
  impulse_ubyte *data;
  printf("Process auth reply\n");
  impulse_message_read(message, "IIxB",
		       &sequence, &phase, &ndata, &data);
  /* WHat the hell do we do now? */
}


static void process_activate_request(impulse_channel channel,
				     impulse_message message)
{
  _p_impulse_channel internal = (_p_impulse_channel)channel->internal;
  impulse_message_formatter formatter[2] = { &version_formatter, &unistring_formatter};
  impulse_unistring *name;
  impulse_uint nversion;
  struct _impulse_channel_version *version = NULL;
  impulse_byte opcode;
  impulse_bool auth;
  impulse_int i;
  impulse_int j;
  impulse_bool found = 0;
  struct _impulse_channel_version rver;
  _p_impulse_channel_service service;
  printf("Process activate reqest\n");
  impulse_message_readf(message, 2, formatter, "rxvBo",
			&name, &nversion, &version, &opcode, &auth);
  impulse_error_raise(IMPULSE_ERROR_WARNING, name[0], NULL);
  printf("Want opcode %d\n", opcode);
  
  /*
   * Check to see if we support the service
   */
  if (!(service = impulse_hash_recover(internal->services, name[0])))
  {
    impulse_error_raise(IMPULSE_ERROR_WARNING, name[0], NULL);
    {
      impulse_iterator it = impulse_hash_keys(internal->services);
      while (!impulse_iterator_eof(it))
      {
	impulse_unistring key = (impulse_unistring)impulse_iterator_next(it);
	impulse_error_raise(IMPULSE_ERROR_WARNING, name[0], key);
      }
      impulse_iterator_destroy(it);
    }
    
    send_error_event(channel, 0, CONTROL_SERVICE_NOT_SUPPORTED,
		     0, CONTROL_ACTIVATE_REQUEST, name[0]);
    goto cleanup;
  }

  /*
   * Check to see if the service is active, unless it has opcode
   * of zero.
   */
  if (opcode &&
      opcode < internal->nincoming &&
      internal->incoming[opcode])
  {
    _p_impulse_channel_service exist = internal->incoming[opcode];
    if (impulse_unistring_equal(exist->external->name, name[0]))
      send_error_event(channel, 0, CONTROL_SERVICE_ACTIVE,
		       0, CONTROL_ACTIVATE_REQUEST, name[0]);
    else
      send_error_event(channel, 0, CONTROL_OPCODE_ACTIVE,
		       0, CONTROL_ACTIVATE_REQUEST, name[0]);
    goto cleanup;
  }
  
  /*
   * Find best version match
   */
  for (i = 0 ; i < nversion ; i++)
  {
    for (j = 0 ; j < service->external->nversion ; j++)
    {
      if (service->external->version[j].major == version[i].major)
      {
	rver.major = version[i].major;
	rver.minor = (version[i].minor > service->external->version[j].minor) ?
	  service->external->version[j].minor : version[i].minor;
	found = 1;
	break;
      }
    }
  }

  /*
   * Damn, we don't support any common versions
   */
  if (!found)
  {
    send_error_event(channel, 0, CONTROL_VERSION_NOT_SUPPORTED,
		     0, CONTROL_ACTIVATE_REQUEST, name[0]);
    goto cleanup;
  }
  
  /*
   * Look for an opcode to use for outgoing messages
   */
  if (opcode)
  {
    found = 0;
    for (i = 0 ; i < internal->noutgoing ; i++)
    {
      if (!internal->outgoing[i])
      {
	service->outgoing = i;
	found = 1;
	break;
      }
    }
    
    if (!found)
    {
      service->outgoing = internal->noutgoing++;
      internal->outgoing = impulse_realloc(internal->outgoing,
					   sizeof(_p_impulse_channel_service)*(internal->noutgoing));
    }
  }
  internal->outgoing[service->outgoing] = service;

  /*
   * Register the incoming opcode
   */
  if (opcode >= internal->nincoming)
  {
    internal->incoming = impulse_realloc(internal->incoming,
					 sizeof(_p_impulse_channel_service)*(opcode+1));
    for(i = internal->nincoming ; i <= opcode ; i++)
      internal->incoming[i] = NULL;
    internal->nincoming = opcode + 1;
  }
  internal->incoming[opcode] = service;
  service->incoming = opcode;
  
  service->active = 1;
  service->auth = 1;
  service->init = 0;

  /*
   * Tell them what we have done!
   */
  printf("Activated service In:%d Out:%d\n", service->incoming, service->outgoing);
  send_activate_reply(channel, 0, opcode, &rver, service->outgoing);

  cleanup:
  impulse_free(version);
  impulse_unistring_destroy(name[0]);
  impulse_free(name);
}


static void process_activate_reply(impulse_channel channel,
				   impulse_message message)
{
  _p_impulse_channel internal = (_p_impulse_channel)channel->internal;
  impulse_message_formatter formatter = &version_formatter;
  impulse_uint sequence;
  struct _impulse_channel_version *version;
  impulse_ubyte opcode; /* Their outgoing opcode (our incoming) */
  impulse_ubyte serviceop; /* Our outgoing opcode */
  _p_impulse_channel_service service;
  printf("Process activate reply\n");
  impulse_message_readf(message, 1, &formatter, "IBvB",
			&sequence, &serviceop, &version, &opcode);
  printf("Vals %d %d %d %d %d\n", sequence, serviceop, version[0].major, version[0].minor,opcode);
  if (opcode &&
      opcode < internal->nincoming &&
      internal->incoming[opcode])
  {
    _p_impulse_channel_service exist = internal->incoming[opcode];
    send_error_event(channel, 0, CONTROL_OPCODE_ACTIVE,
		     0, CONTROL_ACTIVATE_REQUEST, NULL);
  }

  if (opcode >= internal->nincoming)
  {
    impulse_int i;
    internal->incoming = impulse_realloc(internal->incoming,
					 sizeof(_p_impulse_channel_service)*(opcode+1));
    for(i = internal->nincoming ; i <= opcode ; i++)
      internal->incoming[i] = NULL;
    internal->nincoming = opcode + 1;
  }
  service = internal->outgoing[serviceop];
  internal->incoming[opcode] = service;
  service->incoming = opcode;
  service->active = 1;
  service->auth = 1;
  
  printf("Activate confirm Out:%d In:%d\n", serviceop, opcode);
}


static void process_deactivate_request(impulse_channel channel,
				       impulse_message message)
{
  impulse_ubyte opcode;
  printf("Process deactivate request\n");
  impulse_message_read(message, "B",
		       opcode);

  send_deactivate_reply(channel, 0, 1);
}


static void process_deactivate_reply(impulse_channel channel,
				     impulse_message message)
{
  impulse_uint sequence;
  impulse_bool accept;
  printf("PRocess deactivate reply\n");
  impulse_message_read(message, "Io",
		       &sequence, &accept);
  
}


static void process_query_request(impulse_channel channel,
				  impulse_message message)
{
  impulse_message_formatter formatter = &unistring_formatter;
  impulse_unistring *name;
  printf("Process query reqyes\n");
  impulse_message_readf(message, 1, &formatter, "r",
			&name);
}


static void process_query_reply(impulse_channel channel,
				impulse_message message)
{
  impulse_message_formatter formatter = &version_formatter;
  impulse_uint sequence;
  impulse_uint nversion;
  struct _impulse_channel_version *version;
  printf("Process query reply\n");
  impulse_message_readf(message, 1, &formatter, "Ixv",
			sequence, nversion, &version);
  
  impulse_free(version);
}


static void process_ping_request(impulse_channel channel,
				 impulse_message message)
{
  impulse_uint secs;
  impulse_uint usecs;
  printf("Process ping request\n");
  impulse_message_read(message, "II",
		       secs, usecs);
  send_ping_reply(channel, 0, secs, usecs);
}


static void process_ping_reply(impulse_channel channel,
			       impulse_message message)
{
  impulse_uint sequence;
  impulse_uint secs;
  impulse_uint usecs;
  struct timeval tv;
  printf("Ping reply\n");
  impulse_message_read(message, "III",
		       sequence, secs, usecs);
  
  gettimeofday(&tv, NULL);
  
  printf("Got Ping reply from %d:%d at %d:%d\n", secs, usecs, (int)tv.tv_sec, (int)tv.tv_usec);
}


/* End of file */
