/*                                                                |- Impulse -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is the Impulse utility library
 * for Impulse. (Part of the ROWS distribution)
 *
 * The Initial Developer of the Original Code is
 * Daniel Berrange. Portions created by Daniel Berrange
 * are Copyright (C) 1999 Daniel Berrange.
 * All Rights Reserved
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * transport.h
 * -----------
 *
 * Communication transport protocol
 */


#ifndef __impulse_comms_transport_h
#define __impulse_comms_transport_h

#include <impulse/util/unistring.h>
#include <impulse/util/event.h>
#include <impulse/util/thread.h>

/*                            ======--------======                           */

/* 
 * Error values.
 * This should cope with all the expected / recoverable situations. ANy
 * other problems are probably fatal and so will call impulse_error_raise...
 */
#define IMPULSE_TRANSPORT_SUCCESS             0  /* Wahey it worked ! */
#define IMPULSE_TRANSPORT_ADDRESS_UNKNOWN    -1  /* The address was not known */
#define IMPULSE_TRANSPORT_ADDRESS_INVALID    -2  /* The address format was invalid */
#define IMPULSE_TRANSPORT_ADDRESS_INUSE      -3  /* The address is already in use */
#define IMPULSE_TRANSPORT_ADDRESS_UNAVAIL    -4  /* Address resolution failed */
#define IMPULSE_TRANSPORT_ADDRESS_RESOLVE    -5  /* The address is being resolved */
#define IMPULSE_TRANSPORT_CONNECTION_PENDING -6  /* A connection is pending */
#define IMPULSE_TRANSPORT_CONNECTION_RESET   -7  /* The connection was reset */
#define IMPULSE_TRANSPORT_CONNECTION_REJECT  -8  /* The connection was rejected */
#define IMPULSE_TRANSPORT_CONNECTION_TIMEOUT -9  /* THe connection timed out */
#define IMPULSE_TRANSPORT_CONNECTION_FAILED  -10 /* The conection attempt failed */
#define IMPULSE_TRANSPORT_CONNECTION_CLOSED  -11 /* No connection is open */

typedef enum _impulse_transport_event
{
  IMPULSE_TRANSPORT_EVENT_DATA = 0,         /* Incoming data is pending */
  IMPULSE_TRANSPORT_EVENT_CONNECT,          /* The connection has been established */
  IMPULSE_TRANSPORT_EVENT_RESOLVE,          /* The name resolution request has completed */
  IMPULSE_TRANSPORT_EVENT_INCOMING,         /* There is an incoming connection request */
} impulse_transport_event;


typedef struct _impulse_transport_impl *impulse_transport_impl;
typedef struct _impulse_transport *impulse_transport;

typedef impulse_bool (*impulse_transport_handler)(impulse_transport,
						  impulse_transport_event,
						  void *);

struct _impulse_transport_impl
{
  /* Creation stuff */
  void (*create)(impulse_transport);
  void (*destroy)(impulse_transport);

  /* Event handling stuff */
  impulse_int (*handle)(impulse_transport,
			impulse_transport_handler,
			void *);
  impulse_int (*unhandle)(impulse_transport);
  
  /* Name resolution stuff */
  impulse_int (*lookup)(impulse_transport,
			impulse_unistring,
			impulse_unistring *);
  impulse_int (*rlookup)(impulse_transport,
			 impulse_unistring,
			 impulse_unistring *);
  
  /* Connection management */
  impulse_int (*open)(impulse_transport,
		      impulse_unistring,
		      impulse_unistring);
  impulse_int (*accept)(impulse_transport,
			impulse_transport*);
  impulse_int (*close)(impulse_transport);
  impulse_int (*active)(impulse_transport);

  /* Data transfer stuff */
  impulse_int (*pending)(impulse_transport);
  impulse_int (*send)(impulse_transport,
		      impulse_byte *,
		      impulse_uint,
		      impulse_unistring);
  impulse_int (*receive)(impulse_transport,
			 impulse_byte *,
			 impulse_uint,
			 impulse_unistring *);
};

struct _impulse_transport
{
  impulse_transport_impl impl;
  impulse_unistring protocol;
  impulse_unistring local;
  impulse_unistring remote;
#if USE_THREADS
  impulse_mutex lock;
#endif
  void *internal;
};


impulse_transport impulse_transport_create(impulse_unistring protocol);
void impulse_transport_destroy(impulse_transport transport);

/* Convenience macros */

/* Mutual exclusion */

#if USE_THREADS
#define impulse_transport_lock(t) impulse_mutex_lock(t->lock)
#define impulse_transport_unlock(t) impulse_mutex_unlock(t->lock)
#else
#define impulse_transport_lock(t)
#define impulse_transport_unlock(t)
#endif

/* Event macros */
#define impulse_transport_handle(t,f,d) t->impl->handle(t,f,d)
#define impulse_transport_unhandle(t) t->impl->unhandle(t)

/* Name resolution */
#define impulse_transport_lookup(t,n,a) t->impl->lookup(t,n,a)
#define impulse_transport_rlookup(t,a,n) t->impl->rlookup(t,a,n)

/* Initiation macros */
#define impulse_transport_open(t,l,r) t->impl->open(t,l,r)
#define impulse_transport_close(t) t->impl->close(t)
#define impulse_transport_active(t) t->impl->active(t)
#define impulse_transport_accept(t,i) t->impl->accept(t,i)

/* Data transfer functions */
#define impulse_transport_pending(t) t->impl->pending(t)
#define impulse_transport_send(t,b,s,a) t->impl->send(t,b,s,a)
#define impulse_transport_receive(t,b,s,a) t->impl->receive(t,b,s,a)

#endif

/* End of file */
