/*                                                                  |- CFSFC -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is Copyright (C) 1998 Andrew Hunter.
 * All Rights Reserved. 
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter.
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * child.h
 * -------
 *
 * Something contained by a container
 */

#ifndef __impulse_rows_child_h
#define __impulse_rows_child_h

#include <impulse/rows/eventbox.h>
#include <impulse/rows/drawable.h>
#include <impulse/rows/brush.h>
#include <impulse/rows/input.h>

#ifdef __cplusplus
extern "C" {  
#endif

/*                            ======--------======                           */

/* Contained widget type */

typedef void **ROWS_child;

#include <impulse/rows/container.h>
/*                            ======--------======                           */


/* Events */


typedef enum {
  ROWS_CHILD_PREMOVE,     /* Before the actual movement takes place */
  ROWS_CHILD_PRESIZE,     /* Before the actual sizing takes place */
  ROWS_CHILD_MOVE,        /* After the actual movement has taken place */
  ROWS_CHILD_SIZE,        /* After the actual sizing has taken place */
  
  ROWS_CHILD_KEY,         /* Key press & release events */
  ROWS_CHILD_INPUT,       /* Typing & clicking Input events */
  
  ROWS_CHILD_BUTTON,      /* Button press & release events */
  ROWS_CHILD_MOTION,      /* Pointer motion events */
  ROWS_CHILD_CROSSING,    /* Pointer enter / exit events */
  
  ROWS_CHILD_REALISATION,  /* When a widget is realised */
  ROWS_CHILD_EXPOSURE,     /* When part of a widget needs repainting */
  ROWS_CHILD_VISIBILITY,   /* When a widget is shown / hidden */
  /* This one will probably change / disappear soon */
  ROWS_CHILD_STATE,        /* State stuff */

  ROWS_CHILD_EVENT_MAX
} ROWS_child_event_num;


typedef struct _ROWS_child_event_move
{
  ROWS_point position;
} ROWS_child_event_move;


typedef struct _ROWS_child_event_size
{
  ROWS_point size;
} ROWS_child_event_size;


typedef struct _ROWS_child_event_key
{
  impulse_bool pressed;
  impulse_unichar key;
  impulse_int state;
} ROWS_child_event_key;


typedef struct _ROWS_child_event_input
{
  impulse_unichar key;
  impulse_int state;
} ROWS_child_event_input;


typedef struct _ROWS_child_event_button
{
  impulse_bool pressed;
  impulse_int button;
  impulse_int state;
  ROWS_point position;
} ROWS_child_event_button;


typedef struct _ROWS_child_event_crossing
{
  impulse_bool entered;
  impulse_int state;
  ROWS_point position;
} ROWS_child_event_crossing;


typedef struct _ROWS_child_event_motion
{
  impulse_int state;
  ROWS_point position;
} ROWS_child_event_motion;


typedef struct _ROWS_child_event_realisation
{
  impulse_bool realised;
} ROWS_child_event_realisation;


typedef struct _ROWS_child_event_visibility
{
  impulse_bool visible;
} ROWS_child_event_visibility;


typedef struct _ROWS_child_event_exposure
{
  ROWS_rect clip;
} ROWS_child_event_exposure;


#if 1
/* Fix me: do this better */
typedef enum _ROWS_child_state
{
  ROWS_CHILD_FOCUSED,
  ROWS_CHILD_UNFOCUSED,
  ROWS_CHILD_SELECTED,
  ROWS_CHILD_UNSELECTED,
  ROWS_CHILD_SHADED
} ROWS_child_state;

typedef enum _ROWS_child_grouping
{
  ROWS_CHILD_PRESSED    = 0x10000,
  ROWS_CHILD_INPUTFOCUS
} ROWS_child_grouping;

typedef struct _ROWS_child_event_changed {
  ROWS_child_state new_state;
  impulse_int            grouping;
  
  ROWS_widget      new_focus;
} ROWS_child_event_changed;
#endif


typedef union _ROWS_child_event
{
  struct _ROWS_child_event_move move;
  struct _ROWS_child_event_size size;
  struct _ROWS_child_event_key key;
  struct _ROWS_child_event_input input;
  struct _ROWS_child_event_button button;
  struct _ROWS_child_event_crossing crossing;
  struct _ROWS_child_event_motion motion;
  struct _ROWS_child_event_realisation realisation;
  struct _ROWS_child_event_visibility visibility;
  struct _ROWS_child_event_exposure exposure;
  struct _ROWS_child_event_changed changed;
} ROWS_child_event;


/* Functions */

typedef impulse_bool (*ROWS_child_func) (ROWS_child widget,
					 ROWS_child_event_num type,
					 ROWS_child_event *event,
					 void *data);

/*                            ======--------======                           */

/* Standard widget functions */

extern ROWS_widget       ROWS_child_create         (ROWS_drawable drawable);
extern ROWS_widget      *ROWS_child_initialise     (int id,
						    ROWS_drawable drawable);
extern inline int        ROWS_child_next_identifier(void);
extern inline ROWS_child ROWS_child_cast           (void **widget);

/*                            ======--------======                           */

/* Class-specific functions */

/* 
 * Note: to prevent some very tricky deadlock situations, use this function
 *       to lock child widgets, NOT ROWS_widget_lock. (Well, by tricky, I
 *       mean it took me an afternoon to spot it even with help :-)
 *
 *       ROWS_child_lock locks this child and its containers, and its
 *       containers containers. It is generally always safe to call this.
 *       Under some circumstances, it is safe just to use ROWS_widget_lock
 *       on the child widget. (If you're only manipulating the child widget
 *       and not going to touch its container). Be cautious of mutex
 *       notifications under debugging (disable LinuxThreads extensions),
 *       as these could be the first signs of trouble. (You lock the child
 *       widget, do something that affects the container, ROWS_container
 *       calls ROWS_child_lock, which relocks the child, producing a mutex
 *       recursion notice)
 */

/* Locking functions */
extern void ROWS_child_lock  (ROWS_child child);
extern void ROWS_child_unlock(ROWS_child child);

/* Event stuff */
extern void ROWS_child_event_register(ROWS_child child,
				      ROWS_child_event_num type,
				      ROWS_child_func func,
				      void *data);
extern void ROWS_child_event_deregister(ROWS_child child,
					ROWS_child_event_num type,
					ROWS_child_func func,
					void *data);
extern void ROWS_child_event_send(ROWS_child child,
				  ROWS_child_event_num type,
				  ROWS_child_event *event);
extern impulse_bool ROWS_child_event_is_handled(ROWS_child child,
						ROWS_child_event_num type);


/* Size / location functions */
extern void ROWS_child_set_size(ROWS_child child,
				ROWS_point *size);
extern inline ROWS_point *ROWS_child_get_size(ROWS_child child);

extern void ROWS_child_set_border(ROWS_child child,
				  ROWS_point *border);
extern ROWS_point *ROWS_child_get_border(ROWS_child child);

extern void ROWS_child_set_origin(ROWS_child child,
				  ROWS_point *origin);
extern ROWS_point *ROWS_child_get_origin(ROWS_child child);
extern ROWS_point *ROWS_child_get_offset(ROWS_child child);


/* Functions to do with viewing and creating widgets */
extern void ROWS_child_set_visible(ROWS_child child,
				   impulse_bool visible);
extern impulse_bool ROWS_child_is_visible(ROWS_child child);

extern void ROWS_child_set_realised(ROWS_child child,
				    impulse_bool realised);
extern impulse_bool ROWS_child_is_realised(ROWS_child child);

extern void ROWS_child_set_actualised(ROWS_child child,
				      impulse_bool actual);
extern impulse_bool ROWS_child_is_actualised(ROWS_child child);

/* Stuff todo with painting */
extern ROWS_container ROWS_child_get_container(ROWS_child child);
extern ROWS_drawable ROWS_child_get_drawable(ROWS_child child);

extern inline ROWS_style ROWS_child_get_style(ROWS_child child);

extern void ROWS_child_update(ROWS_child child);


/* Miscellaneous state functions */

extern void ROWS_child_set_focus(ROWS_child child,
				 impulse_bool focused);
extern impulse_bool ROWS_child_has_focus(ROWS_child child);


#if 1
/* 
 * Fix me: find a better way of doing what this does.
 * This involves redefining how the whole ROWS_style malarky works.
 */
extern void ROWS_child_set_activation(ROWS_child             child,
				      ROWS_style_active_type type);
extern void ROWS_child_set_state     (ROWS_child             child,
				      ROWS_child_state       state,
				      impulse_int            grouping);

extern ROWS_child_state       ROWS_child_get_state     (ROWS_child       child);
extern ROWS_style_active_type ROWS_child_get_activation(ROWS_child       child);
extern ROWS_child             ROWS_child_get_selected  (ROWS_child_state state);
#endif

#ifdef __cplusplus
}
#endif

#endif
