/*                                                                  |- CFSFC -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is the ROWS widget library
 * for Impulse.
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter. Portions created by Andrew Hunter
 * are Copyright (C) 1999 Andrew Hunter.
 * All Rights Reserved
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * colour.c
 * --------
 *
 * ROWS colour manipulation
 */

#include "../../../config.h"

#include <stdio.h>
#include <math.h>

#include "impulse/rows/main.h"
#include "impulse/rows/drawable_int.h"
#include "impulse/util/memory.h"
#include "impulse/util/thread.h"
#include "impulse/util/error.h"

extern ROWS_colour ROWS_colour_cie_for_drawable(void **drawable,
						int x,
						int y,
						int z)
{
  ROWS_colour_desc col;

  col.type = ROWS_colour_XYZ;
  col.data.xyz.x = x;
  col.data.xyz.y = y;
  col.data.xyz.z = z;
  
  return ROWS_colour_for_drawable(drawable, col, NULL, 0);
}

extern ROWS_colour ROWS_colour_rgb_for_drawable(void **drawable,
						int r,
						int g,
						int b)
{
  ROWS_colour_desc col;

  col.type = ROWS_colour_RGB;
  col.data.rgb.r = r;
  col.data.rgb.g = g;
  col.data.rgb.b = b;
  
  return ROWS_colour_for_drawable(drawable, col, NULL, 0);
}

extern ROWS_colour ROWS_colour_for_drawable(void             **drawable,
					    ROWS_colour_desc   colour,
					    ROWS_gamut         gamut,
					    impulse_bool       make_spot)
{
  _p_ROWS_drawable_data **private;
  ROWS_drawable_fparam param;

  ROWS_colour_desc xyz, rgb;
  ROWS_colour      col;

  ROWS_widget_lock(drawable);

  private = (_p_ROWS_drawable_data **) drawable;

  if (gamut == NULL)
    {
      param.colour.drawable = drawable;
      ((*private)->abstraction_layer.function[ROWS_DRAWABLE_COLOUR_GET_GAMUT])
	(&param);
      gamut = param.colour.data.get_gamut.gamut;
    }

  xyz = ROWS_colour_convert(colour, ROWS_colour_XYZ, gamut);
  rgb = ROWS_colour_convert(colour, ROWS_colour_RGB, gamut);

  col = impulse_alloc(sizeof(struct _ROWS_colour));

  col->drawable = drawable;
  
  col->x = xyz.data.xyz.x;
  col->y = xyz.data.xyz.y;
  col->z = xyz.data.xyz.z;

  col->r = rgb.data.rgb.r;
  col->g = rgb.data.rgb.g;
  col->b = rgb.data.rgb.b;

  col->gamut   = gamut;
  col->is_spot = make_spot;
  
  param.colour.drawable = drawable;
  param.colour.data.create.xyz = xyz;
  param.colour.data.create.rgb = rgb;
  param.colour.data.create.is_spot = 
  param.colour.data.create.data = NULL;
  ((*private)->abstraction_layer.function[ROWS_DRAWABLE_COLOUR_CREATE])
    (&param);
  col->data = param.colour.data.create.data;
  
  ROWS_widget_unlock(drawable);

  return col;
}

/*
extern ROWS_colour ROWS_colour_create_for_drawable(void **drawable,
						   int    r,
						   int    g,
						   int    b)
{
  ROWS_colour col;
  _p_ROWS_drawable_data **private;
  ROWS_drawable_fparam param;

  ROWS_widget_lock(drawable);

  private = (_p_ROWS_drawable_data **) drawable;

  col = impulse_alloc(sizeof(struct _ROWS_colour));

  col->drawable = drawable;
  col->r    = r;
  col->g    = g;
  col->b    = b;

  param.colour.drawable = drawable;
  param.colour.data.create.r = r;
  param.colour.data.create.g = g;
  param.colour.data.create.b = b;

  ((*private)->abstraction_layer.function[ROWS_DRAWABLE_COLOUR_CREATE])
    (&param);
  col->data = param.colour.data.create.data;

  ROWS_widget_unlock(drawable);

  return col;
}
*/

extern ROWS_colour ROWS_colour_create_pixel_colour(void **drawable,
						   impulse_int  pixel)
{
  ROWS_colour col;
  _p_ROWS_drawable_data **private;
  ROWS_drawable_fparam param;

  ROWS_widget_lock(drawable);

  private = (_p_ROWS_drawable_data **) drawable;

  col = impulse_alloc(sizeof(struct _ROWS_colour));

  col->drawable = drawable;
  col->r    = -1;
  col->g    = -1;
  col->b    = -1;

  param.colour.drawable                = drawable;
  param.colour.data.create_pixel.pixel = pixel;

  ((*private)->abstraction_layer.function[ROWS_DRAWABLE_COLOUR_PIXEL])
    (&param);
  col->data = param.colour.data.create.data;

  ROWS_widget_unlock(drawable);

  return col;
}

void ROWS_colour_destroy(ROWS_colour c)
{
  _p_ROWS_drawable_data **private;
  ROWS_drawable_fparam param;

  ROWS_widget_lock(c->drawable);

  private = (_p_ROWS_drawable_data **) c->drawable;

  param.colour.drawable          = c->drawable;
  param.colour.data.discard.data = c->data;

  ((*private)->abstraction_layer.function[ROWS_DRAWABLE_COLOUR_DISCARD])
    (&param);

  ROWS_widget_unlock(c->drawable);

  impulse_free(c);
}

/*                            ======--------======                           */

/* Colour space functions */

#define NEED_GAMUT { if (gamut==NULL) { ROWS_error_raise(IMPULSE_ERROR_CRITICAL, ROWS_ERROR_INTERNAL, 0, "Colour conversion requires RGB gamut"); } }
ROWS_colour_desc ROWS_colour_convert(ROWS_colour_desc from,
				     ROWS_colour_type to,
				     ROWS_gamut       gamut)
{
  ROWS_colour_desc result;
  
  if (from.type == to)
    {
      return from;
    }

  switch (from.type)
    {
    case ROWS_colour_HSV:
    case ROWS_colour_HLS:
    case ROWS_colour_CMYK:
      /* Fixme: These colours are easier converted via RGB */
      break;

      /* == Colorimetric schemes == */
    case ROWS_colour_LAB:
      {
	ROWS_colour_desc xyz;
	double x,y,z;
	double Xn, Yn, Zn;

	double fYYn; /* f(Y/Yn) */
	double alim;

	NEED_GAMUT; /* We need to know where white is */

	/* Coordinates of white (0-1) */
	Xn = (gamut->C[0] * gamut->r[0] +
	      gamut->C[1] * gamut->g[0] +
	      gamut->C[2] * gamut->b[0]) /16777216.0;
	Yn = (gamut->C[0] * gamut->r[1] +
	      gamut->C[1] * gamut->g[1] +
	      gamut->C[2] * gamut->b[1]) /16777216.0;
	Zn = ((4096.0-gamut->r[0] - gamut->r[1]) * gamut->C[0] +
	      (4096.0-gamut->g[0] - gamut->g[1]) * gamut->C[1] +
	      (4096.0-gamut->b[0] - gamut->b[1]) * gamut->C[2])
	  /16777216.0;
	
	xyz.type = ROWS_colour_XYZ;
	if (from.data.lab.l < 328 /* 0.08 */)
	  {
	    y = (from.data.lab.l/(9.033*4096.0)) * Yn;
	    fYYn = 7.787*(y/Yn)+(16.0/116.0);
	  }
	else
	  {
	    y = Yn *
	      pow((25.0/(29.0*4096.0))*from.data.lab.l +
		  4.0/29.0, 3);
	    fYYn = pow(y/Yn, 1.0/3.0);
	  }

	/* Point at which X/Xn becomes <0.008856 */
	alim = (1.03446517212 - 5 * fYYn)*4096.0;
	if (from.data.lab.a < alim)
	  {
	    x = (Xn*((20.0/29.0)-(from.data.lab.a/4096.0)))/(-38.935);
	  }
	else
	  {
	    x = (pow(5.0*fYYn + (from.data.lab.a/4096.0), 3)*Xn)/125;
	  }

	/* Z/Zn is never <0.008856 */
	z = (pow(2.0*fYYn-(from.data.lab.b/4096.0), 3)*Zn)/8.0;

	xyz.data.xyz.x = x*4096.0;
	xyz.data.xyz.y = y*4096.0;
	xyz.data.xyz.z = z*4096.0;
	
	return ROWS_colour_convert(xyz, to, gamut);
      }
      break;

    case ROWS_colour_LUV:
      {
	double Xn, Yn, Zn;
	double YYn;
	double Upn, Vpn; /* u'n, v'n... */
	double X, Y, Z;
	double L, u, v;
	
	ROWS_colour_desc xyz;
	
	xyz.type = ROWS_colour_XYZ;

	L = from.data.luv.l/4096.0;
	u = from.data.luv.u/4096.0;
	v = from.data.luv.v/4096.0;
	
	/* Coordinates of white (0-1) */
	Xn = (gamut->C[0] * gamut->r[0] +
	      gamut->C[1] * gamut->g[0] +
	      gamut->C[2] * gamut->b[0]) /16777216.0;
	Yn = (gamut->C[0] * gamut->r[1] +
	      gamut->C[1] * gamut->g[1] +
	      gamut->C[2] * gamut->b[1]) /16777216.0;
	Zn = ((4096.0-gamut->r[0] - gamut->r[1]) * gamut->C[0] +
	      (4096.0-gamut->g[0] - gamut->g[1]) * gamut->C[1] +
	      (4096.0-gamut->b[0] - gamut->b[1]) * gamut->C[2])
	  /16777216.0;

	YYn = pow((25.0/29.0)*L+(4.0/29.0), 3.0);
	if (YYn>0.008856)
	  {
	    Y = YYn*Yn;
	  }
	else
	  {
	    Y = (100.0/903.0)*L*Yn;
	  }

	Upn = 4.0*Xn/(Xn+15.0*Yn+3.0*Zn);
	Vpn = 9.0*Yn/(Xn+15.0*Yn+3.0*Zn);

	/* Here be large equations :-/ */
	X =
	  ((117.0*L*Y)/(13.0*L*Vpn+v)) -
	  15.0*Y +
	  (((3.0/4.0) * (20.0*L*Y +
			 ((39.0*Upn*Y*L*L)/(13.0*L*Vpn+v)) +
			 ((3.0*L*Y*u)/(13.0*L*Vpn+v)) -
			 ((156.0*Y*L*L)/(13.0*L*Vpn+v)))) /
	   L);

	Z =
	  (((156.0*Y*L*L)/(13.0*L*Vpn+v)) -
	   ((3.0*L*Y*u)/(13.0*L*Vpn+v)) -
	   ((39.0*Upn*Y*L*L)/(13.0*L*Vpn+v)) -
	   (20.0*L*Y)) /
	  (4.0*L);

	xyz.data.xyz.x = X*4096.0;
	xyz.data.xyz.y = Y*4096.0;
	xyz.data.xyz.z = Z*4096.0;

	return ROWS_colour_convert(xyz, to, gamut);
      }
      break;
      
    case ROWS_colour_YUV:
      {
	ROWS_colour_desc xyz;
	double u,v;
	double x,y; /* Yxy coordinates. Confusing, huh? */
	double X,Y,Z;
	
	xyz.type = ROWS_colour_XYZ;
	
	xyz.data.xyz.y = from.data.yuv.y;
	Y = from.data.yuv.y/4096.0;
	u = from.data.yuv.u/4096.0;
	v = from.data.yuv.v/4096.0;

	x =
	  (6.0/((u/v)+(2.0/v)-4.0)) -
	  (9.0/(v*((3.0*u/v)+(6.0/v) - 12.0))) +
	  1.5;
	y =
	  1.0/
	  ((u/v) + (2.0/v) - 4.0);

	X = (Y*x)/y;
	Z = (Y/y) - ((Y*x)/y) - Y;

	xyz.data.xyz.x = X*4096.0;
	xyz.data.xyz.z = Z*4096.0;
	
	return ROWS_colour_convert(xyz, to, gamut);
      }
      break;
      
    case ROWS_colour_YUpVp:
      {
	ROWS_colour_desc xyz;
	double u,v;
	double x,y; /* Yxy coordinates. Confusing, huh? */
	double X,Y,Z;
	
	xyz.type = ROWS_colour_XYZ;
	
	xyz.data.xyz.y = from.data.yupvp.y;
	Y = from.data.yupvp.y/4096.0;
	u = from.data.yupvp.up/4096.0;
	v = from.data.yupvp.vp/4096.0;

	x =
	  (12.0 /
	   (3.0*((u/v)+(2.0/v))-8.0)) -
	  (27 /
	   (v*(((9.0*u)/v)+(18.0/v)-24.0))) +
	  (3.0/2.0);
	y = 2/
	  (((3.0*u)/v)+(6.0/v)-8.0);

	X = (Y*x)/y;
	Z = (Y/y) - ((Y*x)/y) - Y;

	xyz.data.xyz.x = X*4096.0;
	xyz.data.xyz.z = Z*4096.0;
	
	return ROWS_colour_convert(xyz, to, gamut);
      }
      break;

    case ROWS_colour_Yxy:
      {
	ROWS_colour_desc xyz;
	int Y;

	xyz.type = ROWS_colour_XYZ;
	Y = xyz.data.xyz.y = from.data.Yxy.Y;
	
	xyz.data.xyz.x = (Y*from.data.Yxy.x)/from.data.Yxy.y;
	xyz.data.xyz.z = ((Y*4096)/from.data.Yxy.y) -
	  ((Y*from.data.Yxy.x)/from.data.Yxy.y) - Y;
	
	return ROWS_colour_convert(xyz, to, gamut);
      }
      break;
      
    case ROWS_colour_Chs:
    case ROWS_colour_Chab:
      /* Fixme: These colours are easier converted via XYZ */
      break;
      
    default:
      break;
    }

  switch (to)
    {
    case ROWS_colour_HSV:
    case ROWS_colour_HLS:
    case ROWS_colour_CMYK:
      /* Fixme: These colours are easier converted via RGB */
      break;

    case ROWS_colour_LUV:
    case ROWS_colour_YUV:
    case ROWS_colour_YUpVp:
    case ROWS_colour_Chs:
    case ROWS_colour_LAB:
    case ROWS_colour_Chab:
      /* Fixme: These colours are easier converted via XYZ */
      break;

    default:
      break;
    }
  
  if (from.type == ROWS_colour_XYZ)
    {
      result.type = to;
      
      switch (to)
	{
	case ROWS_colour_RGB:
	  NEED_GAMUT;

	  result.data.rgb.r =
	    from.data.xyz.x*gamut->invmatrix[0][0]+
	    from.data.xyz.y*gamut->invmatrix[0][1]+
	    from.data.xyz.z*gamut->invmatrix[0][2];

	  result.data.rgb.g =
	    from.data.xyz.x*gamut->invmatrix[1][0]+
	    from.data.xyz.y*gamut->invmatrix[1][1]+
	    from.data.xyz.z*gamut->invmatrix[1][2];

	  result.data.rgb.b =
	    from.data.xyz.x*gamut->invmatrix[2][0]+
	    from.data.xyz.y*gamut->invmatrix[2][1]+
	    from.data.xyz.z*gamut->invmatrix[2][2];

	  result.data.rgb.r = (result.data.rgb.r>>14);
	  result.data.rgb.g = (result.data.rgb.g>>14);
	  result.data.rgb.b = (result.data.rgb.b>>14);

	  if (gamut->gamma.r != 1.0)
	    {
	      result.data.rgb.r =
		pow(result.data.rgb.r/gamut->gamma.Kr, gamut->gamma.ir);
	    }
	  if (gamut->gamma.g != 1.0)
	    {
	      result.data.rgb.g =
		pow(result.data.rgb.g/gamut->gamma.Kg, gamut->gamma.ig);
	    }
	  if (gamut->gamma.b != 1.0)
	    {
	      result.data.rgb.b =
		pow(result.data.rgb.b/gamut->gamma.Kb, gamut->gamma.ib);
	    }

	  /* RGB values are pegged (not a great approximation) */
	  if (result.data.rgb.r < 0)
	    result.data.rgb.r = -1;
	  if (result.data.rgb.g < 0)
	    result.data.rgb.g = -1;
	  if (result.data.rgb.b < 0)
	    result.data.rgb.b = -1;
	  
	  if (result.data.rgb.r > 4095)
	    result.data.rgb.r = 4096;
	  if (result.data.rgb.g > 4095)
	    result.data.rgb.g = 4096;
	  if (result.data.rgb.b > 4095)
	    result.data.rgb.b = 4096;
	  break;

	default:
	  ROWS_error_raise(IMPULSE_ERROR_CRITICAL,
			   ROWS_ERROR_INTERNAL,
			   0,
			   "Colour model not supported");
	  break;
	}
    }
  else if (to == ROWS_colour_XYZ)
    {
      result.type = to;
      
      switch (from.type)
	{
	case ROWS_colour_RGB:
	  NEED_GAMUT;

	  if (gamut->gamma.r != 1.0)
	    {
	      from.data.rgb.r =
		gamut->gamma.Kr*pow(from.data.rgb.r, gamut->gamma.r);
	    }
	  if (gamut->gamma.g != 1.0)
	    {
	      from.data.rgb.g =
		gamut->gamma.Kg*pow(from.data.rgb.g, gamut->gamma.g);
	    }
	  if (gamut->gamma.b != 1.0)
	    {
	      from.data.rgb.b =
		gamut->gamma.Kb*pow(from.data.rgb.b, gamut->gamma.b);
	    }

	  result.data.xyz.x =
	    from.data.rgb.r*gamut->matrix[0][0]+
	    from.data.rgb.g*gamut->matrix[0][1]+
	    from.data.rgb.b*gamut->matrix[0][2];

	  result.data.xyz.y =
	    from.data.rgb.r*gamut->matrix[1][0]+
	    from.data.rgb.g*gamut->matrix[1][1]+
	    from.data.rgb.b*gamut->matrix[1][2];

	  result.data.xyz.z =
	    from.data.rgb.r*gamut->matrix[2][0]+
	    from.data.rgb.g*gamut->matrix[2][1]+
	    from.data.rgb.b*gamut->matrix[2][2];

	  result.data.xyz.x = result.data.xyz.x>>14;
	  result.data.xyz.y = result.data.xyz.y>>14;
	  result.data.xyz.z = result.data.xyz.z>>14;
	  break;

	default:
	  ROWS_error_raise(IMPULSE_ERROR_CRITICAL,
			   ROWS_ERROR_INTERNAL,
			   0,
			   "Colour model not supported");
	}
    }
  else
    {
      return ROWS_colour_convert(ROWS_colour_convert(from,
						     ROWS_colour_XYZ,
						     gamut),
				 to,
				 gamut);
    }

  return result;
}

ROWS_gamut ROWS_colour_make_gamut (int Rx, int Ry,
				   int Gx, int Gy,
				   int Bx, int By,
				   int Cr,
				   int Cg,
				   int Cb,
				   double gamma)
{
  double matrix[3][3];
  double invmatrix[3][3];
  float  r[2], g[2], b[2];
  float  c[3];
  
  ROWS_gamut result;

  /* Hmm, yeuch, I think */
  r[0] = Rx/4096.0; r[1] = Ry/4096.0;
  g[0] = Gx/4096.0; g[1] = Gy/4096.0;
  b[0] = Bx/4096.0; b[1] = By/4096.0;
  c[0] = Cr/4096.0; c[1] = Cg/4096.0; c[2] = Cb/4096.0;

  /* [X,Y,Z] = matrix * [ R, G, B ] */
  matrix[0][0] = r[0]*c[0];
  matrix[0][1] = g[0]*c[1];
  matrix[0][2] = b[0]*c[2];
  
  matrix[1][0] = r[1]*c[0];
  matrix[1][1] = g[1]*c[1];
  matrix[1][2] = b[1]*c[2];

  matrix[2][0] = (1-r[0]-r[1])*c[0];
  matrix[2][1] = (1-g[0]-g[1])*c[1];
  matrix[2][2] = (1-b[0]-b[1])*c[2];

  /* Work out the inverse matrix ([R,G,B] = invmatrix * [X,Y,Z]) */
  {
    double factor;

    factor =
      matrix[0][0]*matrix[1][1]*matrix[2][2] +
      matrix[0][1]*matrix[1][2]*matrix[2][0] +
      matrix[0][2]*matrix[1][0]*matrix[2][1] -
      matrix[0][2]*matrix[1][1]*matrix[2][0] -
      matrix[0][0]*matrix[1][2]*matrix[2][1] -
      matrix[0][1]*matrix[1][0]*matrix[2][2];

    invmatrix[0][0] =
      (matrix[2][2]*matrix[1][1] - matrix[1][2]*matrix[2][1]) /
      factor;
    invmatrix[0][1] =
      (matrix[0][2]*matrix[2][1] - matrix[2][2]*matrix[0][1]) /
      factor;
    invmatrix[0][2] =
      (matrix[1][2]*matrix[0][1] - matrix[0][2]*matrix[1][1]) /
      factor;

    invmatrix[1][0] =
      (matrix[1][2]*matrix[2][0] - matrix[2][2]*matrix[1][0]) /
      factor;
    invmatrix[1][1] =
      (matrix[2][2]*matrix[0][0] - matrix[0][2]*matrix[2][0]) /
      factor;
    invmatrix[1][2] =
      (matrix[0][2]*matrix[1][0] - matrix[1][2]*matrix[0][0]) /
      factor;

    invmatrix[2][0] =
      (matrix[2][1]*matrix[1][0] - matrix[1][1]*matrix[2][0]) /
      factor;
    invmatrix[2][1] =
      (matrix[0][1]*matrix[2][0] - matrix[2][1]*matrix[0][0]) /
      factor;
    invmatrix[2][2] =
      (matrix[1][1]*matrix[0][0] - matrix[0][1]*matrix[1][0]) /
      factor;
  }

  /* Fill in the important fields */
  result = impulse_alloc(sizeof(struct _ROWS_gamut));

  {
    int x,y;

    for (x=0; x<3; x++)
      for (y=0; y<3; y++)
	{
	  result->matrix[x][y]    = matrix[x][y]*16384;
	  result->invmatrix[x][y] = invmatrix[x][y]*16384;
	}
  }

  /* Fill in the rest of the fields */
  result->r[0] = Rx; result->r[1] = Ry;
  result->g[0] = Gx; result->g[1] = Gy;
  result->b[0] = Bx; result->b[1] = By;

  result->C[0] = Cr;
  result->C[1] = Cg;
  result->C[2] = Cb;

  result->gamma.r    =
    result->gamma.g  =
    result->gamma.b  = gamma;
  result->gamma.ir   =
    result->gamma.ig =
    result->gamma.ib = 1.0/gamma;
  result->gamma.Kr   =
    result->gamma.Kg =
    result->gamma.Kb = 4095/(pow(4095, gamma));

  result->gamma.t_r   =
    result->gamma.t_g =
    result->gamma.t_b = NULL;

  result->rgb_func    = NULL;
  result->xyz_func    = NULL;
  result->custom_func = NULL;
  result->func_data   = NULL;
  
  return result;
}

/* End of file */
