/*                                                                |- Impulse -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is the ROWS widget library
 * for Impulse.
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter. Portions created by Andrew Hunter
 * are Copyright (C) 1999 Andrew Hunter.
 * All Rights Reserved
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * ROWS_Xcomm.c
 * --------
 *
 * ROWS X-specific communications routines
 */

#include "impulse/rows/comms_int.h"
#include "impulse/util/memory.h"
#include "impulse/util/error.h"
#include "impulse/util/thread.h"
#include "impulse/util/event.h"
#include "impulse/rows/main_xwin.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <X11/Xlib.h>

/*                            ======--------======                           */

static impulse_int comms_autobind = 0;
static Window comms_window;
static impulse_sparse comms_events = NULL;
#if USE_THREADS
static impulse_mutex_static comms_lock = IMPULSE_MUTEX_CREATE_STATIC;
#endif


/*                            ======--------======                           */

struct _ROWS_comms_header
{
  impulse_uint window;
  impulse_uint atom;
  impulse_uint length;
};

struct _ROWS_comms_raw
{
  impulse_uint length;
  impulse_ubyte *data;
  impulse_ubyte *offset;
  
  struct _ROWS_comms_raw *next;
};

struct _ROWS_comms_data
{
  struct _ROWS_comms_header header;
  impulse_ubyte *data;
  
  struct _ROWS_comms_data *next;
};

struct _ROWS_comms
{
  impulse_unistring local;
  impulse_unistring remote;
  
  Window window_local;
  Window window_remote;
  Atom atom_local;
  Atom atom_remote;

  impulse_bool deleted;
  impulse_bool exists;
  struct _ROWS_comms_data *head, *tail;
  
  impulse_transport_handler handler;
  void *data;
};

typedef struct _ROWS_comms_header *ROWS_comms_header;
typedef struct _ROWS_comms_raw *ROWS_comms_raw;
typedef struct _ROWS_comms_data *ROWS_comms_data;
typedef struct _ROWS_comms *ROWS_comms;

/*                            ======--------======                           */
static void process_property(ROWS_comms comms,
			     ROWS_comms_raw raw)
{
  ROWS_comms_data current = NULL;
  struct _ROWS_comms_header header;
  impulse_int need = sizeof(struct _ROWS_comms_header);
  impulse_ubyte *dest = (impulse_ubyte *)&header;
  impulse_bool body = 0;

  while (raw)
  {
    if (raw->length <= need)
    {
      ROWS_comms_raw temp;
      memcpy(dest, raw->data, raw->length);
      need -= raw->length;
      dest += raw->length;
      
      temp = raw->next;
      XFree(raw->data);
      impulse_free(raw);
      raw = temp;
    }
    else
    {
      memcpy(dest, raw->data, need);
      memmove(raw->data, raw->data + need, raw->length-need);
      raw->length -= need;
      need = 0;
    }
    
    if (!need)
    {
      if (body)
      {
	dest = (impulse_ubyte *)&header;

	need = sizeof(struct _ROWS_comms_header);
	body = 0;
	
	if (comms->head)
	  comms->tail->next = current;
	else
	  comms->head = current;
	comms->tail = current;
      }
      else
      {
	current = impulse_alloc(sizeof(struct _ROWS_comms_data));
	current->header.window = impulse_uint_be(header.window);
	current->header.atom = impulse_uint_be(header.atom);
	current->header.length = impulse_uint_be(header.length);
	current->next = NULL;
	
	need = current->header.length;
	current->data = impulse_alloc(sizeof(impulse_ubyte)*need);
	dest = current->data;
	body = 1;
      }
    }
  } 

  /* 
   * We must be expecting the first byte of the next header,
   * otherwise bad shit has happened. ie. Like someone screwed
   * around with ROWS_comms_send so that it no longer appends
   * the header and body all in one operation
   */
  if (dest != (impulse_ubyte *)&header)
  {
    impulse_unistring desc = impulse_unistring_create_utf("ROWS communications got screwed up (partial property found)", 8);
    impulse_error_raise(IMPULSE_ERROR_CRITICAL, desc, NULL);
  }
}


static impulse_bool transfer_property(ROWS_comms comms)
{
  ROWS_comms_raw head = NULL, tail = NULL;
  Atom return_type;
  impulse_int return_format;
  impulse_int bytes_got;
  impulse_int bytes_remain;
  impulse_uint offset = 0;
  impulse_uint length = 3;
  impulse_uint iter = 0;

  do
  {
    impulse_char *data;
    ROWS_comms_raw raw = NULL;
    XGetWindowProperty(appl->display,
		       comms->window_local,
		       comms->atom_local,
		       offset, length,
		       True,
		       comms->atom_local,
		       &return_type,
		       &return_format,
		       (unsigned long *)&bytes_got,
		       (unsigned long *)&bytes_remain,
		       (unsigned char **)&data);

    raw = impulse_alloc(sizeof(struct _ROWS_comms_raw));
    raw->data = data;
    raw->length = bytes_got;
    raw->next = NULL;
    
    if (head)
      tail->next = raw;
    else
      head = raw;
    tail = raw;
    
    if (bytes_remain)
    {
      offset = length;
      length = bytes_remain;
      iter++;
      
      /*
       * If another client is sending data faster than we can
       * read it, grab the server so we can process all thats
       * currently waiting
       */
      if (iter == 5)
	XGrabServer(appl->display);
    }    
  } while (bytes_remain);

  if (iter >= 5)
    XUngrabServer(appl->display);
  comms->deleted = 1;
  comms->exists = 0;
  
  process_property(comms, head);
  return 1;
}


static impulse_bool window_event(impulse_event *event,
				 void *data)
{
  XEvent *xevent = (XEvent *)event->data.system.data;
  impulse_bool ret = 0;
  
#if USE_THREADS
//  impulse_mutex_lock(&comms_lock);
#endif

  switch (xevent->type)
  {
   case PropertyNotify:
    {
      impulse_transport transport = (impulse_transport)impulse_sparse_recover(comms_events, xevent->xproperty.atom);
      ROWS_comms comms = (ROWS_comms)transport->internal;
      
      if (xevent->xproperty.state == PropertyNewValue)
      {
	if (!comms || !comms->handler || comms->deleted)
	{
	  break;
	}
	if (!xevent->xproperty.send_event)
	  comms->exists = 1;

	ret = (*comms->handler)(transport, IMPULSE_TRANSPORT_EVENT_DATA, comms->data);

	/* 
	 * If there is still some data waiting to be collected then
	 * retrigger the event. We do this indirectly, so that we
	 * don't monopolise the event loop with badly written clients.
	 */
	if (comms->head)
	{
	  XSendEvent(appl->display, comms_window, False, None, xevent);
	}
      } 
      else if (xevent->xproperty.state == PropertyDelete) 
      {
	comms->deleted = 0;
	comms->exists = 0;
      }
    }
    break;
    
   default:
    break;
  }
  impulse_free(xevent);
#if USE_THREADS
//  impulse_mutex_unlock(&comms_lock);
#endif
  
  return 1;
}


static impulse_bool ROWS_comms_split(impulse_unistring address,
				     Window *window,
				     Atom *atom)
{
  impulse_int offset1 = impulse_unistring_find_chr(address, (impulse_unichar)':');
  impulse_int offset2 = impulse_unistring_rfind_chr(address, (impulse_unichar)':');
  impulse_unistring chunk;
  impulse_char *start;
  impulse_char *end;
  
  if (offset1 < 0 ||
      offset2 < 0 ||
      offset1 == offset2)
    return 0;

  chunk = impulse_unistring_create_space(0);
  if ((offset2 - offset1) > 1)
  {    
    impulse_unistring_substring(chunk, address, offset1+1, offset2);
    
    start = impulse_unistring_to_utf(chunk, 8, 0);
    *window = strtol(start, &end, 10);
    if (start == end)
      *window = 0;

    impulse_free(start);
  }
  else
    *window = 0;
  
  if ((address->length - offset2) > 1)
  {
    impulse_unistring_substring(chunk, address, offset2+1, address->length);
    
    start = impulse_unistring_to_utf(chunk, 8, 0);
    *atom = strtol(start, &end, 10);
    if (start == end)
      *atom = 0;

    impulse_free(start);
  }
  else
    *atom = 0;
  
  impulse_unistring_destroy(chunk);
  return 1;
}


static impulse_bool ROWS_comms_join(Window window,
				    Atom atom,
				    impulse_unistring *address)
{
  impulse_char buffer[1024];

  sprintf(buffer, "win:%d:%d", (impulse_uint)window, (impulse_uint)atom);
  *address = impulse_unistring_create_utf(buffer, 8);
  
  return 1;
}

		    
void ROWS_comms_init(void)
{
#if USE_THREADS
  impulse_mutex_lock(&comms_lock);
#endif

  comms_window = XCreateWindow(appl->display,
			       appl->screen_window,
			       0, 0,
			       1, 1,
			       0,
			       0,
			       InputOnly,
			       CopyFromParent,
			       0,
			       NULL);

  XSelectInput(appl->display,
	       comms_window,
	       PropertyChangeMask);
  
  impulse_event_register(IMPULSE_EVENT_GUI,
			 comms_window,
			 window_event,
			 NULL);

  comms_events = impulse_sparse_create();
  
#if USE_THREADS
  impulse_mutex_unlock(&comms_lock);
#endif
}


void ROWS_comms_cleanup(void)
{
#if USE_THREADS
  impulse_mutex_lock(&comms_lock);
#endif

  impulse_sparse_destroy(comms_events);
  impulse_event_deregister(IMPULSE_EVENT_GUI,
			   comms_window,
			   NULL);
  
  XDestroyWindow(appl->display, comms_window);
  
#if USE_THREADS
  impulse_mutex_unlock(&comms_lock);
#endif
}


void ROWS_comms_create(impulse_transport transport)
{
  ROWS_comms comms = impulse_alloc(sizeof(struct _ROWS_comms));

  comms->window_local = None;
  comms->window_remote = None;
  comms->atom_local = None;
  comms->atom_remote = None;

  comms->deleted = 0;
  comms->exists = 0;
  comms->head = NULL;
  comms->tail = NULL;
  comms->handler = NULL;
  comms->data = NULL;
  
  transport->internal = comms;
}


void ROWS_comms_destroy(impulse_transport transport)
{
  ROWS_comms comms = (ROWS_comms)transport->internal;

  if (comms->window_local)
    ROWS_comms_close(transport);
  
  impulse_free(comms);
}


/* Event functions */
impulse_int ROWS_comms_handle(impulse_transport transport,
			      impulse_transport_handler func,
			      void *edata)
{
  ROWS_comms comms = (ROWS_comms)transport->internal;
  
#if USE_THREADS
  impulse_mutex_unlock(&comms_lock);
#endif

  comms->handler = func;
  comms->data = edata;

  /* Fix me: maintain a list of handled comms stuff,
   * and only add us when we have a handler. THis
   * will protect us if new data appears while we
   * are changing this in another thread. */

  if (comms->atom_local)
    impulse_sparse_store(comms_events, comms->atom_local, transport);
  
#if USE_THREADS
  impulse_mutex_unlock(&comms_lock);
#endif

  return 0;
}


impulse_int ROWS_comms_unhandle(impulse_transport transport)
{
  ROWS_comms comms = (ROWS_comms)transport->internal;

#if USE_THREADS
  impulse_mutex_lock(&comms_lock);
#endif
  
  if (comms->atom_local)
    impulse_sparse_remove(comms_events, comms->atom_local);
  comms->handler = NULL;
  comms->data = NULL;
  
#if USE_THREADS
  impulse_mutex_unlock(&comms_lock);
#endif
  
  return 0;
}


/* Name resolution stuff */

impulse_int ROWS_comms_lookup(impulse_transport transport,
                              impulse_unistring name,
                              impulse_unistring *address)
{
  *address = impulse_unistring_create(name);
  return 0;
#if 0
  /* 
   * Fix me: finish code to turn atom names into values
   */
  impulse_int offset = impulse_unistring_find_chr(name, (impulse_unichar)':');
  impulse_unistring atomid;
  impulse_unistring atomname;
  impulse_char *
  Atom atom;
  
  if (offset < 0)
    return IMPULSE_TRANSPORT_ADDRESS_INVALID;
  
  atomid = impulse_unistring_create_sub(name, offset+1, name->length);
#endif 
}


impulse_int ROWS_comms_rlookup(impulse_transport transport,
			       impulse_unistring address,
			       impulse_unistring *name)
{
  *name = impulse_unistring_create(address);
  return 0;
}


/* Connection functions */
impulse_int ROWS_comms_open(impulse_transport transport,
			    impulse_unistring local,
			    impulse_unistring remote)
{
  ROWS_comms comms = (ROWS_comms)transport->internal;
  impulse_int ret = 0;
  
#if USE_THREADS
  impulse_mutex_lock(&comms_lock);
#endif
  if (local)
  {
    if ((ret = ROWS_comms_split(local,
				&comms->window_local,
				&comms->atom_local)) < 0)
      return ret;
  }
  comms->window_local = comms_window;
  if (!comms->atom_local)
  {
    comms->atom_local = comms_autobind++;
    ROWS_comms_join(comms->window_local, comms->atom_local, &transport->local);
  }

  if (remote)
  {
    if ((ret = ROWS_comms_split(remote,
				&comms->window_remote,
				&comms->atom_remote)) < 0)
    {
      comms->window_local = 0;
      comms->atom_local = 0;
      return ret;
    }
    transport->remote = impulse_unistring_create(remote);
  }
  else
    transport->remote = NULL;

  comms->deleted = 0;
  comms->exists = 0;
  
#if USE_THREADS
  impulse_mutex_unlock(&comms_lock);
#endif
  return 0;
}


impulse_int ROWS_comms_accept(impulse_transport transport,
			      impulse_transport *client)
{
  ROWS_comms comms = (ROWS_comms)transport->internal;
#if USE_THREADS
  impulse_mutex_lock(&comms_lock);
#endif

#if USE_THREADS
  impulse_mutex_unlock(&comms_lock);
#endif
  
  /* 
   * This is just a hack to stop the compiler complaining
   * about an unused variable
   */
  if (comms)
    ;

  /* 
   * We should probably generate an error or warning 
   * message here
   */
  return 0;
}


impulse_int ROWS_comms_close(impulse_transport transport)
{
  ROWS_comms comms = (ROWS_comms)transport->internal;
  ROWS_comms_data current, next = NULL;
  
#if USE_THREADS
  impulse_mutex_lock(&comms_lock);
#endif
  
  if (comms->local)
    impulse_unistring_destroy(comms->local);
  if (comms->remote)
    impulse_unistring_destroy(comms->remote);
  comms->local = NULL;
  comms->remote = NULL;
  comms->window_local = None;
  comms->window_remote = None;
  comms->atom_local = None;
  comms->atom_remote = None;
  comms->deleted = 0;
  comms->exists = 0;
  
  current = comms->head;
  while (current)
  {
    next = current->next;
    impulse_free(current->data);
    impulse_free(current);
    current = next;
  }
  comms->head = NULL;
  comms->tail = NULL;

#if USE_THREADS
  impulse_mutex_unlock(&comms_lock);
#endif
  return 0;
}


impulse_int ROWS_comms_active(impulse_transport transport)
{
  ROWS_comms comms = (ROWS_comms)transport->internal;
  
  return comms->window_local != None;
}


/* Data transfer functions */


impulse_int ROWS_comms_pending(impulse_transport transport)
{
  ROWS_comms comms = (ROWS_comms)transport->internal;
  ROWS_comms_data current;
  impulse_uint count = 0;
  
#if USE_THREADS
  impulse_mutex_lock(&comms_lock);
#endif
  if (comms->exists)
    transfer_property(comms);
#if USE_THREADS
  impulse_mutex_unlock(&comms_lock);
#endif

  current = comms->head;
  while (current)
  {
    count += current->header.length;
    current = current->next;
  }
  
  return count;
}


impulse_int ROWS_comms_send(impulse_transport transport,
			    impulse_byte *buffer,
			    impulse_uint size,
			    impulse_unistring address)
{
  ROWS_comms comms = (ROWS_comms)transport->internal;
  ROWS_comms_header header;
  Window window_send;
  Atom atom_send;
  
  if (address)
  {
    if (!ROWS_comms_split(address,
			  &window_send,
			  &atom_send))
      return -1;
  }
  else
  {
    window_send = comms->window_remote;
    atom_send = comms->atom_remote;
  }

  header = impulse_alloc(sizeof(struct _ROWS_comms_header)+size);
  header->window = impulse_uint_be(comms->window_local);
  header->atom = impulse_uint_be(comms->atom_local);
  header->length = impulse_uint_be(size);
  memcpy(((char *)header) + sizeof(struct _ROWS_comms_header), buffer, size);

  /*
   * Note we stick the *entire* datagram (ie the header and body) onto the
   * property all at once. This ensures that the client reading the property
   * cannot get a property containing a fraction of a datagram.
   */
  ROWS_X_expecting_error(1);
  XChangeProperty(appl->display,
		  window_send,
		  atom_send,
		  atom_send,
		  8,
		  PropModeAppend,
		  (char *)header,
		  sizeof(struct _ROWS_comms_header)+size);
  ROWS_X_expecting_error(0);
  
  return 0;
}


impulse_int ROWS_comms_receive(impulse_transport transport,
			       impulse_byte *buffer,
			       impulse_uint size,
			       impulse_unistring *address)
{
  ROWS_comms comms = (ROWS_comms)transport->internal;
  ROWS_comms_data pdata;
  impulse_int got;
  
#if USE_THREADS
  impulse_mutex_lock(&comms_lock);
#endif
  if (comms->exists)
    transfer_property(comms);
#if USE_THREADS
  impulse_mutex_unlock(&comms_lock);
#endif

  if (!comms->head)
  {
    if (address)
      *address = NULL;
    return 0;
  }
  
  pdata = comms->head;
  comms->head = pdata->next;
  if (!comms->head)
    comms->tail = NULL;
  
  got = (size > pdata->header.length) ? pdata->header.length : size;
  
  memcpy(buffer, pdata->data, got);
  
  if (address)
    ROWS_comms_join(pdata->header.window, pdata->header.atom, address);

  impulse_free(pdata->data);
  impulse_free(pdata);

  return got;
}

/* End of file */
