/*                                                                  |- CFSFC -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is Copyright (C) 1998 Andrew Hunter.
 * All Rights Reserved. 
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter.
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * constructor.h
 * -------------
 *
 * A widget used for constructing complex user interface artifacts :)
 * (Or, put more simply, Charles's guideline widget idea)
 */

#include "../../../config.h"

#include "impulse/rows/main.h"
#include "impulse/rows/constructor_int.h"
#include "impulse/rows/container_int.h"
#include "impulse/rows/child_int.h"
#include "impulse/util/memory.h"

/*                            ======--------======                           */

static void ROWS_constructor_setup(ROWS_widget *widget);
static int  constructor_ID   = -1;
static int  constructor_type = -1;

static void ROWS_constructor_destroy(void **private);

/* Child events */
static impulse_bool _p_ROWS_constructor_redraw(ROWS_child widget,
					       ROWS_child_event_num event_number,
					       ROWS_child_event *event,
					       void *data);
static impulse_bool _p_ROWS_constructor_resize(ROWS_child widget,
					       ROWS_child_event_num event_number,
					       ROWS_child_event *event,
					       void *data);

static impulse_bool _p_ROWS_constructor_layout(ROWS_child widget,
					       ROWS_child_event_num event_number,
					       ROWS_child_event *event,
					       void *data);


/*                            ======--------======                           */

/* Standard widget functions */

ROWS_widget ROWS_constructor_create(void)
{
  ROWS_widget *wid;

  if (constructor_ID   == -1)
    constructor_ID   = ROWS_container_next_identifier();

  wid = ROWS_constructor_initialise(constructor_ID);

  return *wid;
}


ROWS_widget *ROWS_constructor_initialise(int id)
{
  ROWS_widget *widget;

  if (constructor_ID   == -1)
    constructor_ID   = ROWS_container_next_identifier();
  if (constructor_type == -1)
    constructor_type = ROWS_widget_next_type();

  widget = ROWS_container_initialise(id, NULL);
  ROWS_widget_set_type(widget, constructor_ID, constructor_type);
  ROWS_widget_set_destructor(widget, constructor_ID, ROWS_constructor_destroy);

  ROWS_constructor_setup(widget);

  return widget;
}


int ROWS_constructor_next_identifier(void)
{
  if (constructor_ID==-1)
    constructor_ID = ROWS_container_next_identifier();

  return constructor_ID + 1;
}


ROWS_constructor ROWS_constructor_cast(void **widget)
{
  if (constructor_ID   == -1)
    constructor_ID   = ROWS_container_next_identifier();
  if (constructor_type == -1)
    constructor_type = ROWS_widget_next_type();

  return ROWS_widget_cast_misc(widget, constructor_ID, constructor_type);
}


void ROWS_constructor_setup(ROWS_widget *widget)
{
  _p_ROWS_constructor_data **private;

  private  = (_p_ROWS_constructor_data **) ROWS_widget_private(widget, constructor_ID);
  *private = impulse_alloc(sizeof(_p_ROWS_constructor_data));

  (*private)->array         = NULL;
  (*private)->no_guidelines = -1;

  /* GLs 0 & 1 are the left & right guidelines */
  ROWS_constructor_add_GL(ROWS_constructor_cast(*widget),
			  ROWS_CONSTRUCTOR_VERTICAL,
			  ROWS_CONSTRUCTOR_MIN,
			  -1, -1, 0);
  ROWS_constructor_add_GL(ROWS_constructor_cast(*widget),
			  ROWS_CONSTRUCTOR_VERTICAL,
			  ROWS_CONSTRUCTOR_MAX,
			  -1, -1, 0);

  /* GLs 2 & 3 are the top & bottom guidelines */
  ROWS_constructor_add_GL(ROWS_constructor_cast(*widget),
			  ROWS_CONSTRUCTOR_HORIZONTAL,
			  ROWS_CONSTRUCTOR_MIN,
			  -1, -1, 0);
  ROWS_constructor_add_GL(ROWS_constructor_cast(*widget),
			  ROWS_CONSTRUCTOR_HORIZONTAL,
			  ROWS_CONSTRUCTOR_MAX,
			  -1, -1, 0);

  /* This should be removed in non-debugging situations */
  ROWS_child_event_register(ROWS_child_cast(*widget),
			    ROWS_CHILD_EXPOSURE,
			    _p_ROWS_constructor_redraw,
			    NULL);

  ROWS_child_event_register(ROWS_child_cast(*widget),
			    ROWS_CHILD_SIZE,
			    _p_ROWS_constructor_resize,
			    NULL);
}


void ROWS_constructor_destroy(void **data)
{
  _p_ROWS_constructor_data  **private;
  _p_ROWS_constructor_child  *next, *last;

  private = (_p_ROWS_constructor_data **) data;

  next = (*private)->children;
  while (next!=NULL)
    {
      ROWS_widget_destroy(next->widget);
      next = next->next;
    }

  while (next!=NULL)
    {
      last = next;
      next = next->next;
      impulse_free(last);
    }
  
  if ((*private)->array!=NULL)
    impulse_free((*private)->array);
  impulse_free(*private);
}

/*                            ======--------======                           */

/* Class-specific functions (guidelines) */

static int ROWS_constructor_add_GL_general(ROWS_constructor             widget,
					   ROWS_constructor_orientation orient,
					   ROWS_constructor_type        type,
					   int   lower_guideline,
					   int   upper_guideline,
					   float position,
					   ROWS_constructor_func  func,
					   void                  *data)
{
  _p_ROWS_constructor_data **private;
  int no_guidelines; /* Help save on RSI... */

  ROWS_widget_lock(widget);

  ROWS_widget_check_type(widget, constructor_type, "ROWS_constructor");

  private = (_p_ROWS_constructor_data **) widget;

  (*private)->no_guidelines++;
  (*private)->array = impulse_realloc((*private)->array,
				   sizeof(_p_ROWS_guideline)*
				   ((*private)->no_guidelines+1));

  /* FIXME: This checking stuff is a bit of a mess */
  if (type == ROWS_CONSTRUCTOR_RELATIVE && lower_guideline == upper_guideline)
      {
	ROWS_error_raise(IMPULSE_ERROR_WARNING, CONSTRUCTOR_GL_INVALID, 0, NULL);
	ROWS_widget_unlock(widget);
        return -1;
      }

  if (type == ROWS_CONSTRUCTOR_RELATIVE
      && (   lower_guideline >  (*private)->no_guidelines
	  || upper_guideline >  (*private)->no_guidelines
	  || lower_guideline <  0
	  || upper_guideline <  0))
      ROWS_error_raise(IMPULSE_ERROR_CRITICAL, CONSTRUCTOR_GL_NONEXISTANT, 0, NULL);

  if (type == ROWS_CONSTRUCTOR_RELATIVE
      && ((*private)->array[lower_guideline].orient != (*private)->array[upper_guideline].orient
	  || ((*private)->array[lower_guideline].orient != orient)))
    ROWS_error_raise(IMPULSE_ERROR_CRITICAL, CONSTRUCTOR_GL_INVALID, 0, NULL);

  no_guidelines = (*private)->no_guidelines;

  (*private)->array[no_guidelines].position         = 0;
  (*private)->array[no_guidelines].position_unknown = 1;
  (*private)->array[no_guidelines].loop_check       = 0;

  (*private)->array[no_guidelines].orient = orient;
  (*private)->array[no_guidelines].type   = type;
  (*private)->array[no_guidelines].lower  = lower_guideline;
  (*private)->array[no_guidelines].upper  = upper_guideline;
  (*private)->array[no_guidelines].pos    = position;

  (*private)->array[no_guidelines].func      = func;
  (*private)->array[no_guidelines].func_data = data;

  ROWS_widget_unlock(widget);

  return (*private)->no_guidelines;
}

int ROWS_constructor_add_custom (ROWS_constructor             widget,
				 ROWS_constructor_orientation orient,
				 int                          lower,
				 int                          upper,
				 ROWS_constructor_func        function,
				 void                        *data)
{
  return
    ROWS_constructor_add_GL_general(widget,
				    orient,
				    ROWS_CONSTRUCTOR_CUSTOM,
				    lower,
				    upper,
				    0.0,
				    function,
				    data); 
}

int ROWS_constructor_add_GL(ROWS_constructor             widget,
			    ROWS_constructor_orientation orient,
			    ROWS_constructor_type        type,
			    int   lower_guideline,
			    int   upper_guideline,
			    float position)
{
  if (type == ROWS_CONSTRUCTOR_CUSTOM)
    {
      impulse_error_raise(IMPULSE_ERROR_CRITICAL,
			  impulse_unistring_create_utf("Use ROWS_constructor_add_custom to create custom guidelines",
						       8),
			  NULL);
    }

  return
    ROWS_constructor_add_GL_general(widget,
				    orient,
				    type,
				    lower_guideline,
				    upper_guideline,
				    position,
				    NULL,
				    NULL);
}

microns ROWS_constructor_GL_position(ROWS_constructor widget,
				     int              guideline)
{
  _p_ROWS_constructor_data **private;
  ROWS_point                *size;
  ROWS_point                *pos;
  microns                    lower, upper;

  ROWS_widget_check_type(widget, constructor_type, "ROWS_constructor");

  private = (_p_ROWS_constructor_data **) widget;

  if ((*private)->array[guideline].loop_check)
      ROWS_error_raise(IMPULSE_ERROR_CRITICAL, CONSTRUCTOR_GL_LOOP, 0, NULL);

  if (   guideline > (*private)->no_guidelines
      || guideline < 0)
    ROWS_error_raise(IMPULSE_ERROR_CRITICAL, CONSTRUCTOR_GL_NONEXISTANT, 0, NULL);

  if (!(*private)->array[guideline].position_unknown)
    {
      return (*private)->array[guideline].position;
    }

  size = ROWS_box_get_virtual (ROWS_box_cast(widget));
  pos  = ROWS_child_get_origin(ROWS_child_cast(widget));

  /* Determine the positions of this guideline */
  (*private)->array[guideline].position_unknown = 0;
  (*private)->array[guideline].loop_check       = 1;

  switch ((*private)->array[guideline].type)
    {
    case ROWS_CONSTRUCTOR_MIN:
      switch ((*private)->array[guideline].orient)
	{
	case ROWS_CONSTRUCTOR_VERTICAL:
	  (*private)->array[guideline].position = 0;/*pos->x;*/
	  break;
	case ROWS_CONSTRUCTOR_HORIZONTAL:
	  (*private)->array[guideline].position = 0;/*pos->y;*/
	  break;
	}
      break;

    case ROWS_CONSTRUCTOR_MAX:
      switch ((*private)->array[guideline].orient)
	{
	case ROWS_CONSTRUCTOR_VERTICAL:
	  (*private)->array[guideline].position = /*pos->x +*/ size->x;
	  break;
	case ROWS_CONSTRUCTOR_HORIZONTAL:
	  (*private)->array[guideline].position = /*pos->y +*/ size->y;
	  break;
	}
      break;

    case ROWS_CONSTRUCTOR_RELATIVE:
      lower = ROWS_constructor_GL_position(widget,
					   (*private)->array[guideline].lower);
      upper = ROWS_constructor_GL_position(widget,
					   (*private)->array[guideline].upper);
      (*private)->array[guideline].position = 
	lower + ((upper-lower)*(*private)->array[guideline].pos);
      break;

    case ROWS_CONSTRUCTOR_ABSOLUTE:
      lower = ROWS_constructor_GL_position(widget,
					   (*private)->array[guideline].lower);
      (*private)->array[guideline].position = 
	lower + (*private)->array[guideline].pos;
      break;

    case ROWS_CONSTRUCTOR_CUSTOM:
      lower = ROWS_constructor_GL_position(widget,
					   (*private)->array[guideline].lower);
      upper = ROWS_constructor_GL_position(widget,
					   (*private)->array[guideline].upper);

      if (lower < upper)
	{
	  microns temp;

	  temp  = lower;
	  lower = upper;
	  upper = temp;
	}

      (*private)->array[guideline].position =
	((*private)->array[guideline].func)(lower,
					    upper,
					    (*private)->array[guideline].func_data);
    }

  (*private)->array[guideline].loop_check = 0;

  return (*private)->array[guideline].position;  
}


ROWS_constructor_orientation ROWS_constructor_GL_orient(ROWS_constructor widget,
							int          guideline)
{
  _p_ROWS_constructor_data **private;

  ROWS_widget_check_type(widget, constructor_type, "ROWS_constructor");

  private = (_p_ROWS_constructor_data **) widget;

  /* FIXME: Need to check for valid values of guideline */
  return (*private)->array[guideline].orient;
}

/*                            ======--------======                           */

/* Event handling functions */

static void _p_resize_widget_func(ROWS_container container,
				  ROWS_child     widget,
				  void           *data,
				  void           *moredata)
{
  ROWS_child_event           *event;
  _p_ROWS_constructor_child  *info;
  ROWS_rect                   rect;
  ROWS_constructor            cons;
  ROWS_child_event            ev;

  ROWS_drawable d1, d2;
  
  ROWS_point origin = *ROWS_child_get_origin(ROWS_child_cast(container));
  
  event = moredata;
  info = data;
  
  cons = ROWS_constructor_cast(container);

  rect.pos.x      = ROWS_constructor_GL_position(cons, info->left);
  rect.pos.y      = ROWS_constructor_GL_position(cons, info->top);
  rect.size.x     = ROWS_constructor_GL_position(cons, info->right)
    - rect.pos.x;
  rect.size.y     = ROWS_constructor_GL_position(cons, info->bottom)
    - rect.pos.y;

  ROWS_box_set_physical(ROWS_box_cast(widget),
			&rect.size);
  ROWS_box_set_position(ROWS_box_cast(widget),
			&rect.pos);
  
  d1 = ROWS_child_get_drawable(ROWS_child_cast(cons));
  d2 = ROWS_child_get_drawable(widget);
  
  if (d1 == d2)
  {
    origin.x += rect.pos.x;
    origin.y += rect.pos.y;
  }  
  
  ROWS_child_set_origin(ROWS_child_cast(widget),
			&origin);

  ev.move.position = rect.pos;
  ROWS_child_event_send(ROWS_child_cast(widget),
			ROWS_CHILD_MOVE,
			&ev);
  
  ev.size.size = rect.size;
  ROWS_child_event_send(ROWS_child_cast(widget),
			ROWS_CHILD_SIZE,
			&ev);
}


static impulse_bool _p_ROWS_constructor_resize(ROWS_child            widget,
					       ROWS_child_event_num  event_number,
					       ROWS_child_event     *event,
					       void                 *data)
{
  _p_ROWS_constructor_data **priv;
  int x;

  ROWS_box_set_virtual(ROWS_box_cast(widget),
		       ROWS_box_get_physical(ROWS_box_cast(widget)));

  /*
   * When we resize, we invalidate all constructor position values.
   * They are recalculated to the new size when ROWS_constructor_GL_position
   * is next called.
   */
  priv = (_p_ROWS_constructor_data **) ROWS_constructor_cast(widget);

  for (x=0; x<=(*priv)->no_guidelines; x++)
    (*priv)->array[x].position_unknown = 1;

  ROWS_container_iter_int(ROWS_container_cast(widget),
			  _p_resize_widget_func,
			  event);

  return 0;
}


static impulse_bool _p_ROWS_constructor_redraw(ROWS_child            widget,
					       ROWS_child_event_num  event_number,
					       ROWS_child_event     *event,
					       void                     *data)
{
  _p_ROWS_constructor_data **private;
  ROWS_drawable              drawable;
  ROWS_style                style;
  /*ROWS_rect                  box;*/
  ROWS_point                *origin;
  ROWS_point                *size;
  int x, pos;

  drawable = ROWS_child_get_drawable(widget);
  private  = (_p_ROWS_constructor_data **) ROWS_constructor_cast(widget);

  style = ROWS_child_get_style(widget);
  /*ROWS_box_get_boundaries(ROWS_box_cast(widget), &box);*/
  origin = ROWS_child_get_offset(widget);
  size   = ROWS_child_get_size(widget);
  
  for (x=0;x<=(*private)->no_guidelines;x++)
    {
      pos = ROWS_constructor_GL_position((ROWS_constructor) private,
					 x);

      switch (ROWS_constructor_GL_orient((ROWS_constructor) private,
					 x))
	{
	case ROWS_CONSTRUCTOR_VERTICAL:
	  ROWS_brush_draw_line(style->brush[ROWS_STYLE_NORMAL][ROWS_STYLE_FOREGROUND],
			       origin->x + pos, origin->y,
			       origin->x + pos, origin->y+size->y);
	  break;
	
	case ROWS_CONSTRUCTOR_HORIZONTAL:
	  ROWS_brush_draw_line(style->brush[ROWS_STYLE_NORMAL][ROWS_STYLE_FOREGROUND],
		       origin->x,         origin->y + pos,
		       origin->x+size->x, origin->y + pos);
	  break;
	}
    }
#if 0
  /* 
   * Fix me: need to get child clipping while drawing parent working
   * properly - currently these two calls just obliterate most of the
   * child!
   */
	    
  ROWS_brush_draw_line(style->brush[ROWS_STYLE_NORMAL][ROWS_STYLE_FOREGROUND],
		       origin->x, origin->y,
		       50000, 50000);
  ROWS_brush_fill_rectangle(style->brush[ROWS_STYLE_HIGHLIGHT][ROWS_STYLE_BACKGROUND],
			    origin->x, origin->y,
			    50000, 50000);
#endif			     
  return 0;
}

/* When we get a changed event, fit the altered widget to its guidelines */
static impulse_bool _p_ROWS_constructor_layout(ROWS_child widget,
					       ROWS_child_event_num event_number,
					       ROWS_child_event *event,
					       void *data)
{
  ROWS_rect                  rect;
  ROWS_constructor           cons = (ROWS_constructor)data;
  _p_ROWS_constructor_child *info;

  info = ROWS_container_lookup_int(ROWS_container_cast(cons), widget);

  if (info)
  {
    ROWS_child_event event;
    
    rect.pos.x  = ROWS_constructor_GL_position(cons, info->left);
    rect.pos.y  = ROWS_constructor_GL_position(cons, info->top);
    rect.size.x = ROWS_constructor_GL_position(cons, info->right)
      - rect.pos.x;
    rect.size.y = ROWS_constructor_GL_position(cons, info->bottom)
      - rect.pos.y;
    
    ROWS_box_set_physical(ROWS_box_cast(widget),
			  &rect.size);
    ROWS_box_set_position(ROWS_box_cast(widget),
			  &rect.pos);
//    ROWS_child_set_origin(ROWS_child_cast(widget),
//			  &rect.pos);
    event.move.position = rect.pos;
    ROWS_child_event_send(widget,
			  ROWS_CHILD_MOVE,
			  &event);
    event.size.size = rect.size;
    ROWS_child_event_send(widget,
			  ROWS_CHILD_SIZE,
			  &event);
  }
  else
  {
    impulse_unistring desc = impulse_unistring_create_utf("Cannot find constructor event information", 8);
    impulse_error_raise(IMPULSE_ERROR_WARNING, desc, NULL);
    impulse_unistring_destroy(desc);
  }

  return 0; /* Maybe someone else wants to know... */
}

/*                            ======--------======                           */

/* Child manipulation functions */

void ROWS_constructor_add_child(ROWS_constructor cons,
				ROWS_child       child,
				int              GL_left,
				int              GL_right,
				int              GL_top,
				int              GL_bottom)
{
  _p_ROWS_constructor_child  *next;

  ROWS_widget_check_type(cons, constructor_type, "ROWS_constructor");

  ROWS_child_lock(ROWS_child_cast(cons));

  next = impulse_alloc(sizeof(_p_ROWS_constructor_child));

  next->widget = child;
  next->top    = GL_top;
  next->bottom = GL_bottom;
  next->left   = GL_left;
  next->right  = GL_right;

  ROWS_child_event_register(child,
			    ROWS_CHILD_PREMOVE,
			    _p_ROWS_constructor_layout,
			    cons);
  ROWS_child_event_register(child,
			    ROWS_CHILD_PRESIZE,
			    _p_ROWS_constructor_layout,
			    cons);
  ROWS_child_event_register(child,
			    ROWS_CHILD_VISIBILITY,
			    _p_ROWS_constructor_layout,
			    cons);
  
  ROWS_container_add_int(ROWS_container_cast(cons),
			 child, 
			 (void *)next);
 
  ROWS_child_unlock(ROWS_child_cast(cons));
}


void ROWS_constructor_remove_child(ROWS_constructor cons,
				   ROWS_child       child)
{
  _p_ROWS_constructor_child  *info;
  
  ROWS_child_lock(ROWS_child_cast(cons));
  
  info = ROWS_container_remove_int(ROWS_container_cast(cons),
				   child);
  impulse_free(info);
  
  ROWS_child_event_deregister(child,
			      ROWS_CHILD_VISIBILITY,
			      _p_ROWS_constructor_layout,
			      cons);
  ROWS_child_event_deregister(child,
			      ROWS_CHILD_PRESIZE,
			      _p_ROWS_constructor_layout,
			      cons);
  ROWS_child_event_deregister(child,
			      ROWS_CHILD_PREMOVE,
			      _p_ROWS_constructor_layout,
			      cons);
  
  ROWS_child_unlock(ROWS_child_cast(cons));
}

				
