/*                                                                |- Impulse -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is the Impulse utility library
 * for Impulse. (Part of the ROWS distribution)
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter. Portions created by Andrew Hunter
 * are Copyright (C) 1999 Andrew Hunter.
 * All Rights Reserved
 *
 * Contributor(s):
 * Portions created by Daniel Berrange
 * are Copyright (C) 1999 Daniel Berrange.
 */

/*
 * main.c
 * ------
 *
 * Impulse program startup and termination functions
 */

#include "../../../config.h"

#include "impulse/util/main.h"
#include "impulse/util/thread.h"
#include "impulse/util/memory.h"
#include "impulse/util/error.h"
#include "impulse/util/event.h"

#include "impulse/util/sparse.h"
#include "impulse/util/hash.h"
#include "impulse/util/list.h"

#include "impulse/util/mapping.h"
#include "impulse/util/text.h"
#include "impulse/util/message.h"

#include <unistd.h>
#include <pwd.h>
#include <grp.h>
#include <locale.h>
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>


/*                            ======--------======                           */


/* Some type declarations */

struct _impulse_app
{
#if USE_THREADS
  impulse_mutex lock;
#endif

  impulse_bool quit;
  
  impulse_unistring names[IMPULSE_APP_NAME_MAX];
  impulse_unistring dirs[IMPULSE_APP_DIR_MAX];
  impulse_unistring command;
  
  impulse_uint argc;
  impulse_unistring *argv;
  
  impulse_hash lopts;
  impulse_sparse sopts;
  impulse_list hooks;
  impulse_list help;
  
  impulse_unistring locale;
  impulse_unistring charset;
  impulse_mapping mapping;
};


struct _impulse_command_details
{
  impulse_unistring lopt;
  impulse_unichar sopt;
  impulse_unistring help;
  impulse_app_arg_type type;
  
  impulse_uint nvalues;
  impulse_unistring *values;
};

typedef struct _impulse_app *impulse_app;
typedef struct _impulse_command_details *impulse_command_details;

static impulse_app app;

/*                            ======--------======                           */

/* Static helper functions */


static void command_cleaner(void *data)
{
  impulse_command_details details = (impulse_command_details)data;
  
  if (details->values)
  {
    impulse_int i;
    for (i = 0 ; i < details->nvalues ; i++)
      impulse_unistring_destroy(details->values[i]);
    impulse_free(details->values);
  }
  
  impulse_unistring_destroy(details->lopt);
  impulse_unistring_destroy(details->help);
  impulse_free(details);
}

static void command_hook(void)
{
  if (impulse_command_argument(_(MSG_CMD_HELP_OPT),NULL))
    impulse_command_help();
  if (impulse_command_argument(_(MSG_CMD_VERBOSE_OPT), NULL))
    impulse_error_handle(IMPULSE_ERROR_INFO, impulse_error_print);
  if (impulse_command_argument(_(MSG_CMD_VERSION_OPT), NULL))
    ;
  if (impulse_command_argument(_(MSG_CMD_DEBUG_OPT), NULL))
  {

  }
}


/*
 * Fix me: internationalise all these messages
 */
#if HAVE_SIGINFO
static void _p_sighandler(int sig, siginfo_t *info, void *data)
#else
static void _p_sighandler(int sig)
#endif
{
  impulse_unistring desc = NULL;
  impulse_unistring msg = NULL;
  impulse_error_level level = IMPULSE_ERROR_EMERGENCY;

  switch (sig)
  {
   case SIGHUP:
    level = IMPULSE_ERROR_CRITICAL;
    desc = _(MSG_SIG_HUP);
    break;
    
   case SIGINT:
    level = IMPULSE_ERROR_CRITICAL;
    desc = _(MSG_SIG_INT);
    break;
    
   case SIGQUIT:
    level = IMPULSE_ERROR_CRITICAL;
    desc = _(MSG_SIG_QUIT);
    break;
    
   case SIGILL:
    level = IMPULSE_ERROR_EMERGENCY;
    desc = _(MSG_SIG_ILL);

#if HAVE_SIGINFO
    switch (info->si_code)
    {
     case ILL_ILLOPC:
      msg = _(MSG_SIG_ILL_OPC);
      break;
     case ILL_ILLOPN:
      msg = _(MSG_SIG_ILL_OPN);
      break;
     case ILL_ILLTRP:
      msg = _(MSG_SIG_ILL_TRAP);
      break;
     case ILL_ILLADR:
      msg = _(MSG_SIG_ILL_ADDR);
      break;
     case ILL_PRVOPC:
      msg = _(MSG_SIG_ILL_PRVOP);
      break;
     case ILL_PRVREG:
      msg = _(MSG_SIG_ILL_PRVREG);
      break;
     case ILL_COPROC:
      msg = _(MSG_SIG_ILL_COPROC);
      break;
     case ILL_BADSTK:
      msg = _(MSG_SIG_ILL_BADSTACK);
      break;
    }
#endif
    break;
    
   case SIGABRT:
    level = IMPULSE_ERROR_EMERGENCY;
    desc = _(MSG_SIG_ABORT);
    break;
    
   case SIGFPE:
    level = IMPULSE_ERROR_CRITICAL;
    desc = _(MSG_SIG_FPE);
    
#if HAVE_SIGINFO
    switch (info->si_code)
    {
     case FPE_INTDIV:
      msg = _(MSG_SIG_FPE_INTDIV);
      break;
     case FPE_INTOVF:
      msg = _(MSG_SIG_FPE_INTOVF);
      break;
     case FPE_FLTDIV:
      msg = _(MSG_SIG_FPE_FLTDIV);
      break;
     case FPE_FLTOVF:
      msg = _(MSG_SIG_FPE_FLTOVF);
      break;
     case FPE_FLTUND:
      msg = _(MSG_SIG_FPE_FLTUND);
      break;
     case FPE_FLTRES:
      msg = _(MSG_SIG_FPE_FLTRES);
      break;
     case FPE_FLTINV:
      msg = _(MSG_SIG_FPE_FLTINV);
      break;
     case FPE_FLTSUB:
      msg = _(MSG_SIG_FPE_FLTSUB);
      break;
    }
#endif
    break;
    
   case SIGSEGV:
    level = IMPULSE_ERROR_EMERGENCY;
    desc = _(MSG_SIG_SEGV);
    
#if HAVE_SIGINFO
    switch(info->si_code)
    {
     case SEGV_MAPERR:
      msg = _(MSG_SIG_SEGV_MAP);
      break;
     case SEGV_ACCERR:
      msg = _(MSG_SIG_SEGV_ACCESS);
      break;
    }
#endif
    break;
    
   case SIGPIPE:
    level = IMPULSE_ERROR_WARNING;
    desc = _(MSG_SIG_PIPE);
    break;
    
   case SIGALRM:
    level = IMPULSE_ERROR_WARNING;
    desc = _(MSG_SIG_ALARM);
    break;
    
   case SIGTERM:
    level = IMPULSE_ERROR_CRITICAL;
    desc = _(MSG_SIG_TERM);
    break;

   case SIGBUS:
    level = IMPULSE_ERROR_EMERGENCY;
    desc = _(MSG_SIG_BUS);
    
#if HAVE_SIGINFO
    switch(info->si_code)
    {
     case BUS_ADRALN:
      msg = _(MSG_SIG_BUS_ALIGN);
      break;
     case BUS_ADRERR:
      msg = _(MSG_SIG_BUS_ADDR);
      break;
     case BUS_OBJERR:
      msg = _(MSG_SIG_BUS_OBJECT);
      break;
    }
#endif
    break;
    
   default:
    level = IMPULSE_ERROR_WARNING;
    desc = _(MSG_SIG_UNEXPECTED);
    break;
  }
  
#if HAVE_SIGINFO
  if (!msg)
  {
    switch(info->si_code)
    {
     case SI_USER:
      msg = _(MSG_SIG_USER);
      break;
     case SI_QUEUE:
      msg = _(MSG_SIG_QUEUE);
      break;
     case SI_TIMER:
      msg = _(MSG_SIG_TIMER);
      break;
     case SI_ASYNCIO:
      msg = _(MSG_SIG_ASYNCIO);
      break;
     case SI_MESGQ:
      msg = _(MSG_SIG_MESGQ);
      break;
    }
  }
#endif
  
  impulse_error_raise(level, desc, msg);
}


#if HAVE_SIGINFO
static void impulse_signal(int signum, void (*signal_func)(int,
							   siginfo_t *,
							   void *))
#else
static void impulse_signal(int signum, void (*signal_func)(int))
#endif
{
  struct sigaction action;
  struct sigaction oldaction;

  sigfillset(&action.sa_mask);
  sigdelset(&action.sa_mask, SIGCHLD);
#if HAVE_SIGINFO
  action.sa_sigaction = signal_func;
  action.sa_flags     = SA_SIGINFO;
#else
  action.sa_handler   = signal_func;
  action.sa_flags     = 0;
#endif
  
  /* 
   * One can't help wondering what will happen when (if) this field is
   * implemented...
   */
  /* action.sa_restorer = NULL; */

  sigaction(signum, &action, &oldaction);
}


typedef struct
{
  impulse_uint len;
  impulse_char *language;
  impulse_char *charset;
} charset_mapper;

static charset_mapper charset_map[] =
{
  /* Official language codes */
  { 2, "ar", "iso88596" },
  { 2, "bg", "iso88595" },
  { 2, "cs", "iso88592" },
  { 2, "cz", "iso88592" },
  { 2, "da", "iso88591" },
  { 2, "de", "iso88591" },
  { 2, "ee", "iso88594" },
  { 2, "el", "iso88597" },
  { 2, "en", "iso88591" },
  { 2, "es", "iso88591" },
  { 2, "et", "iso88594" },
  { 2, "fi", "iso88591" },
  { 2, "fr", "iso88591" },
  { 2, "hr", "iso88592" },
  { 2, "hu", "iso88592" },
  { 2, "is", "iso88591" },
  { 2, "iw", "iso88598" },
  { 2, "ja", "eucjp" },
  { 2, "ko", "euckr" },
  { 2, "lt", "iso88594" },
  { 2, "lv", "iso88594" },
  { 2, "mk", "iso88595" },
  { 2, "nl", "iso88591" },
  { 2, "no", "iso88591" },
  { 2, "pl", "iso88592" },
  { 2, "pt", "iso88591" },
  { 2, "ro", "iso88592" },
  { 2, "ru", "iso88595" },
  { 2, "sh", "iso88592" },
  { 2, "sk", "iso88592" },
  { 2, "sl", "iso88592" },
  { 2, "sp", "iso88595" },
  { 2, "sr", "iso88592" },
  { 2, "sv", "iso88591" },
  { 2, "th", "tactis" },  /* tis620 ??? */
  { 2, "tr", "iso88599" },
  { 4, "zh_TW", "eucTW" },
  { 2, "zh", "eucCN" },
  
  /* Unofficial language names */
  { 8, "american", "iso88591" },
  { 6, "arabic", "iso88596" },
  { 9, "bulgarian", "iso88595" },
  { 8, "c-french", "iso88591" },
  { 9, "chinese-s", "euccn" },
  { 9, "chinese-t", "euctw" },
  { 8, "croatian", "iso88592" },
  { 5, "czech", "iso88592" },
  { 6, "danish", "iso88591" },
  { 5, "dutch", "iso88591" },
  { 7, "english", "iso88591" },
  { 7, "finnish", "iso88591" },
  { 6, "french", "iso88591" },
  { 6, "german", "iso88591" },
  { 5, "greek", "iso88597" },
  { 6, "hebrew", "iso88598" },
  { 9, "hungarian", "iso88592" },
  { 9, "icelandic", "iso88591" },
  { 7, "italian", "iso88591" },
  { 5, "japan", "eucjp" },
  { 6, "korean", "euckr" },
  { 9, "norwegian", "iso88591" },
  { 6, "polish", "iso88592" },
  { 10, "portuguese", "iso88591" },
  { 8, "rumanian", "iso88592" },
  { 7, "russian", "iso88595" },
  { 13, "serbocroatian", "iso88592" },
  { 6, "slovak", "iso88592" },
  { 7, "slovene", "iso88592" },
  { 7, "spanish", "iso88591" },
  { 7, "swedish", "iso88591" },
  { 7, "turkish", "iso88599" },
  { 0, NULL, NULL },
};


/*
 * Why does finding out the correct charset to use have to
 * be such a bitch :(
 */
static impulse_char scratch[1024];
static impulse_char *impulse_charset(impulse_char *locale)
{
  impulse_char *offset;

  if (!(offset = strchr(locale, '.')))
  {
    /* 
     * Search for charset based on language / country code
     */
    charset_mapper *map = charset_map;
    while (map->len)
    {
      if (!strncmp(map->language, locale, map->len))
	return map->charset;
      map++;
    }
  }
  else
  {
    /* 
     * Use embedded charset from locale name
     * Must normalise charset:
     * 1. Remove all character not letters or digits
     * 2. Convert all letters to lowercase
     * 3. Prepend 'iso' if only numbers
     */
    char *dest = scratch;
    char *src = offset+1;
    
    if (strlen(src) > 1022)
      src[1023] = '\0';

    if (isdigit(*src))
    {
      *dest++ = 'i';
      *dest++ = 's';
      *dest++ = 'o';
    }
    while (*src)
    {
      char c = *src++;
      if (isalnum(c))
      {
	if (isupper(c))
	  c = tolower(c);
	*dest++ = c;
      }
    }
    *dest++ = '\0';
    return scratch;
  }
  return "iso88591";
}


/*                            ======--------======                           */

/* Public functions only beyond this point */

void impulse_init(impulse_uint argc,
		  impulse_char **argv,
		  impulse_char *name)
{
  struct passwd *pw;
  struct group *gr;
  impulse_char buffer[1024];
  impulse_char *locale;
  impulse_int i;

#if HAVE_DEBUG_MCHECK
  if (getenv("IMPULSE_MCHECK"))
    impulse_mcheck();
#endif
  app = impulse_alloc(sizeof(struct _impulse_app));

  /* First off lets put in the signal handlers */
  impulse_signal(SIGHUP, _p_sighandler);
  impulse_signal(SIGINT, _p_sighandler);
  impulse_signal(SIGQUIT, _p_sighandler);
  impulse_signal(SIGILL, _p_sighandler);
  impulse_signal(SIGABRT, _p_sighandler);
  impulse_signal(SIGFPE, _p_sighandler);
  impulse_signal(SIGSEGV, _p_sighandler);
  impulse_signal(SIGPIPE, _p_sighandler);
  impulse_signal(SIGALRM, _p_sighandler);
  impulse_signal(SIGTERM, _p_sighandler);
  impulse_signal(SIGBUS, _p_sighandler);

#if USE_THREADS
  app->lock = impulse_mutex_create_recursive();
#endif
  
  pw = getpwuid(getuid());
  gr = getgrgid(getgid());
  gethostname(buffer, 1023);
  app->names[IMPULSE_APP_NAME_PROGRAM] = impulse_unistring_create_utf(name, 8);
  app->names[IMPULSE_APP_NAME_USER] = impulse_unistring_create_utf(pw->pw_name, 8);
  app->names[IMPULSE_APP_NAME_GROUP] = impulse_unistring_create_utf(gr->gr_name, 8);
  app->names[IMPULSE_APP_NAME_PERSON] = impulse_unistring_create_utf(pw->pw_gecos, 8);
  app->names[IMPULSE_APP_NAME_HOST] = impulse_unistring_create_utf(buffer, 8);

  app->dirs[IMPULSE_APP_DIR_HOME] = impulse_unistring_create_utf(pw->pw_dir, 8);
  app->dirs[IMPULSE_APP_DIR_TEMP] = impulse_unistring_create_utf("/tmp", 8);

  app->command = impulse_unistring_create_utf(argv[0], 8);
  
  app->lopts = impulse_hash_create();
  app->sopts = impulse_sparse_create();
  app->hooks = impulse_list_create();
  app->help = impulse_list_create();

  impulse_hash_cleaner(app->lopts, command_cleaner);
  impulse_list_cleaner(app->help, (impulse_list_cleaner_func)impulse_unistring_destroy);
  
  app->argc = argc;
  app->argv = impulse_alloc(sizeof(impulse_unistring)*argc);
  for (i = 0 ; i < argc ; i++)
    app->argv[i] = impulse_unistring_create_utf(argv[i], 8);

  /* Set the actual locale */
  locale = setlocale(LC_ALL, "");
  /* Find out what to use for charset */
  locale = setlocale(LC_CTYPE, NULL);

  /* Default to POSIX */
  if (!locale || !strlen(locale) || !strcmp(locale, "C") || !strcmp(locale, "POSIX"))
    locale = "en";

  app->locale = impulse_unistring_create_utf(locale, 8);
  app->charset = impulse_unistring_create_utf(impulse_charset(locale), 8);
  app->mapping = impulse_mapping_create(app->charset);
  app->quit = 0;
  
  /* Fix me: translate this */
  impulse_command_option(_(MSG_CMD_HELP_OPT), 0, _(MSG_CMD_HELP_MSG), IMPULSE_APP_ARG_NONE);
  impulse_command_option(_(MSG_CMD_VERBOSE_OPT), 0, _(MSG_CMD_VERBOSE_MSG), IMPULSE_APP_ARG_NONE);
  impulse_command_option(_(MSG_CMD_VERSION_OPT), 0, _(MSG_CMD_VERSION_MSG), IMPULSE_APP_ARG_NONE);
  impulse_command_option(_(MSG_CMD_DEBUG_OPT), 0, _(MSG_CMD_DEBUG_MSG), IMPULSE_APP_ARG_NONE);

  impulse_command_hook(command_hook);
  libimputil_text = impulse_text_create(libimputil_text_base,
					impulse_app_locale());
}


void impulse_cleanup(void)
{
  impulse_uint i;
  
  impulse_text_destroy(libimputil_text);
  
  for (i = 0 ; i < app->argc ; i++)
    impulse_unistring_destroy(app->argv[i]);
  impulse_free(app->argv);

  impulse_hash_destroy(app->lopts);
  impulse_sparse_destroy(app->sopts);

  impulse_mapping_destroy(app->mapping);
  impulse_unistring_destroy(app->locale);
  impulse_unistring_destroy(app->charset);
  
  for (i = 0 ; i < IMPULSE_APP_DIR_MAX ; i++)
    impulse_unistring_destroy(app->dirs[i]);
  for (i = 0 ; i < IMPULSE_APP_NAME_MAX ; i++)
    impulse_unistring_destroy(app->names[i]);
  
#if USE_THREADS
  impulse_mutex_destroy(app->lock);
#endif
}


void impulse_main(void)
{
  while (!app->quit)
    impulse_event_handle();
}


void impulse_quit(void)
{
  app->quit = 1;
}


impulse_bool impulse_is_quit(void)
{
  return app->quit;
}


/*                            ======--------======                           */
  
void impulse_command_parse(void)
{
  impulse_int i;
  impulse_int end = app->argc;
  impulse_uint argc = app->argc;
  impulse_unistring *argv = app->argv;
  
#if USE_THREADS
  impulse_mutex_lock(app->lock);
#endif

  for (i = 1 ; i < argc ; i++)
  {
    enum { ARG_LONG, ARG_SHORT, ARG_VAL } type;

    if (argv[i]->length < 2)
      impulse_error_raise(IMPULSE_ERROR_CRITICAL, _(MSG_CMD_MALFORMED), argv[i]);

    /*
     * Three cases to check here:
     * 1. Short arguments start with '-'
     * 2. Long arguments start with '--'
     * 3. Option values are anything else
     */
    if (argv[i]->data[0] != (impulse_unichar)'-')
      type = ARG_VAL;
    else
    {
      if (argv[i]->data[1] != (impulse_unichar)'-')
	type = ARG_SHORT;
      else
	type = ARG_LONG;
    }
    
    /*
     * '--' on its own indicates the end of command arguments,
     * so get the hell out of the loop.
     */
    if ((type == ARG_LONG) &&
	(argv[i]->length == 2))
    {
      if (i < end)
	end = i;
      break;
    }
    
    /*
     * If we have a value without an option, then this could be
     * the first non-option argument, so store is position and continue
     */
    if (type == ARG_VAL)
    {
      if (i < end)
	end = i;
      continue;
    }

    /*
     * If we get this far we must have some form of argument - either
     * long or short. If we have finished processing arguments then
     * this is an error
     */							      
    if (end < i)
      impulse_error_raise(IMPULSE_ERROR_CRITICAL, _(MSG_CMD_MISPLACED), argv[i]);

    /*
     * Either process a long argument
     */
    if (type == ARG_LONG)
    {
      impulse_int offset = impulse_unistring_find_chr(argv[i], (impulse_unichar)'=');
      impulse_unistring arg = NULL;
      impulse_unistring val = NULL;
      impulse_command_details details = NULL;

      /*
       * Chop out the argument name without any leading '--' or trailing embedded
       * value.
       */
      if (offset < 0)
	arg = impulse_unistring_create_sub(argv[i], 2, argv[i]->length);
      else
      {
	arg = impulse_unistring_create_sub(argv[i], 2, offset);
	val = impulse_unistring_create_sub(argv[i], offset+1, argv[i]->length);
      }
      
      /*
       * See if we know about this argument
       */
      details = impulse_hash_recover(app->lopts, arg);
      if (!details)
	impulse_error_raise(IMPULSE_ERROR_CRITICAL, _(MSG_CMD_UNKNOWN), argv[i]);
      
      /*
       * Check to see if argument is expecting a value
       */
      if (details->type == IMPULSE_APP_ARG_NONE)
      {
	if (val)
	  impulse_error_raise(IMPULSE_ERROR_CRITICAL, _(MSG_CMD_MALFORMED), argv[i]);
	details->nvalues++;
      }
      else
      {
	/*
	 * If there was no embedded value then the next argument is taken to be
	 * the value.
	 */
	if (!val)
	{
	  if (++i == argc)
	    impulse_error_raise(IMPULSE_ERROR_CRITICAL, _(MSG_CMD_MALFORMED), argv[i]);
	  if (argv[i]->data[0] == (impulse_unichar)'-')
	    impulse_error_raise(IMPULSE_ERROR_CRITICAL, _(MSG_CMD_MALFORMED), argv[i]);
	  val = impulse_unistring_create(argv[i]);
	}
	/*
	 * Actually parse the value according to the registered type
	 */
	details->values = impulse_realloc(details->values, 
					  sizeof(impulse_unistring) * (details->nvalues+1));
	details->values[details->nvalues++] = impulse_unistring_create(val);
      }
      if (val)
	impulse_unistring_destroy(val);
      if (arg)
	impulse_unistring_destroy(arg);
    }
    else
    {
      /* 
       * Or processs a set of short options
       */
      impulse_int offset = impulse_unistring_find_chr(argv[i], (impulse_unichar)'=');
      impulse_int j;
      
      /*
       * If we have supplied an embedded value, then you cannot group several
       * short arguments together.
       */
      if ((offset != -1) && (offset != 2))
	impulse_error_raise(IMPULSE_ERROR_CRITICAL, _(MSG_CMD_MALFORMED), argv[i]);

      /*
       * Process each of the short arguments in turn
       */
      for (j = 1 ; j < argv[i]->length ; j++)
      {
	impulse_unichar opt = argv[i]->data[j];
	impulse_command_details details = impulse_sparse_recover(app->sopts, opt);

	/*
	 * THis short argument is not known to us
	 */
	if (!details)
	  impulse_error_raise(IMPULSE_ERROR_CRITICAL, _(MSG_CMD_UNKNOWN), argv[i]);
	
	/*
	 * Check to see if we are actually expecting any values for this argument
	 */
	if (details->type == IMPULSE_APP_ARG_NONE)
	{
	  if (offset != -1)
	    impulse_error_raise(IMPULSE_ERROR_CRITICAL, _(MSG_CMD_MALFORMED), argv[i]);
	  details->nvalues++;
	}
	else
	{
	  impulse_unistring val = NULL;
	  /*
	   * Either extract an embedded value, or us the following argument as the value
	   */
	  if (offset != -1)
	  {
	    val = impulse_unistring_create_sub(argv[i], offset+1, argv[i]->length);
	  }
	  else
	  {
	    if (++i == argc)
	      impulse_error_raise(IMPULSE_ERROR_CRITICAL, _(MSG_CMD_MALFORMED), argv[i]);
	    val = impulse_unistring_create(argv[i]);
	  }
	  /* 
	   * Actually process the argument
	   */
	  details->values = impulse_realloc(details->values, 
					    sizeof(impulse_unistring) * (details->nvalues+1));
	  details->values[details->nvalues++] = val;
	}
      }
    }
  }
//  lastarg = end;

  /*
   * Run any post-command parse function hooks
   */
  if (!impulse_list_empty(app->hooks))
  {
    impulse_iterator iterator = impulse_list_elements(app->hooks);
    while (!impulse_iterator_eof(iterator))
    {
      impulse_command_func func = (impulse_command_func)impulse_iterator_next(iterator);
      
      (*func)();
    }
    impulse_iterator_destroy(iterator);
  }

#if USE_THREADS
  impulse_mutex_unlock(app->lock);
#endif
  
  /*
   * That wasn't so difficult now was it!
   */
}


void impulse_command_option(impulse_unistring lopt,
			    impulse_unichar sopt,
			    impulse_unistring help,
			    impulse_app_arg_type type)
{
  impulse_command_details details = NULL;
#if USE_THREADS
  impulse_mutex_lock(app->lock);
#endif

  details = impulse_alloc(sizeof(struct _impulse_command_details));
  details->lopt = impulse_unistring_create(lopt);
  details->sopt = sopt;
  details->type = type;
  details->help = impulse_unistring_create(help);
  details->nvalues = 0;
  details->values = NULL;
  
  impulse_hash_store(app->lopts, details->lopt, details);
  if (sopt)
    impulse_sparse_store(app->sopts, sopt, details);

  /*
   * Fixme: register the argument with the command line help
   */
  
#if USE_THREADS
  impulse_mutex_unlock(app->lock);
#endif
}


impulse_uint impulse_command_argument(impulse_unistring lopt,
				      impulse_unistring **values)
{
  impulse_uint count;
  impulse_command_details details;
  
#if USE_THREADS
  impulse_mutex_lock(app->lock);
#endif
  
  if (!(details = impulse_hash_recover(app->lopts, lopt)))
    impulse_error_raise(IMPULSE_ERROR_CRITICAL, _(MSG_CMD_UNKNOWN), lopt);
  count = details->nvalues;
  if (values)
    *values = details->values;
  
#if USE_THREADS
  impulse_mutex_unlock(app->lock);
#endif
  return count;
}


void impulse_command_hook(impulse_command_func func)
{
#if USE_THREADS
  impulse_mutex_lock(app->lock);
#endif
  impulse_list_append(app->hooks, func);
#if USE_THREADS
  impulse_mutex_unlock(app->lock);
#endif
}


void impulse_command_help(void)
{
  impulse_iterator iterator = NULL;
  
#if USE_THREADS
  impulse_mutex_lock(app->lock);
#endif
  
  iterator = impulse_list_elements(app->help);
  while (!impulse_iterator_eof(iterator))
  {
    impulse_unistring line = (impulse_unistring)impulse_iterator_next(iterator);
    impulse_char *str = impulse_unistring_to_8bit(line, impulse_app_charset());
    
    printf(str);
    impulse_free(str);
  }
  impulse_iterator_destroy(iterator);

#if USE_THREADS
  impulse_mutex_unlock(app->lock);
#endif
}

  
/*                            ======--------======                           */

impulse_unistring impulse_app_dir(impulse_app_dir_type type)
{
  return app->dirs[type];
}


impulse_unistring impulse_app_name(impulse_app_name_type type)
{
  return app->names[type];
}


impulse_unistring impulse_app_command(void)
{
  return app->command;
}


/*                            ======--------======                           */

impulse_unistring impulse_app_locale(void)
{
  return app->locale;
}


impulse_unistring impulse_app_charset(void)
{
  return app->charset;
}


/* End of file */
