/*                                                                |- Impulse -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is Copyright (C) 1998 Andrew Hunter.
 * All Rights Reserved. 
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter.
 *
 * Contributor(s):
 * Portions created by Daniel Berrange
 * are Copyright (C) 1999 Daniel Berrange.
 */

/*
 * mapping.c
 * ---------
 *
 * Implementation of dynamic character set conversions
 */

/*********************************************************************
 * 
 * VIW (Very Important Warning)
 * 
 * No functions used impulse_module_* must *EVER* call
 * 
 *  impulse_unistring_to_8bit
 *  impulse_unistring_from_8bit
 * 
 * Both of these functions make use of dynamically loaded modules,
 * and thus would have huge potential for infinite loops.....
 * 
 ********************************************************************/

#include "../../../config.h"

#include "impulse/util/mapping_int.h"
#include "impulse/util/hash.h"
#include "impulse/util/memory.h"
#include "impulse/util/thread.h"
#include "impulse/util/module.h"
#include "impulse/util/error.h"

#include <stdlib.h>

/*                            ======--------======                           */

static impulse_hash encodings = NULL;
#if USE_THREADS
static impulse_mutex_static encoding_mutex = IMPULSE_MUTEX_CREATE_STATIC;
#endif

struct _impulse_mapping
{
  impulse_unistring encoding;
  impulse_int refs;
  impulse_module module;
  impulse_mapping_engine engine;
};


/*                            ======--------======                           */


static impulse_char *impulse_mapping_to_ascii(impulse_unistring str)
{
  impulse_char *chunk = impulse_alloc(str->length+1);
  impulse_uint i;
  
  for (i = 0 ; i < str->length ; i++)
    chunk[i] = (str->data[i] <= IMPULSE_BYTE_MAX) ? (impulse_char)str->data[i] : '?';
  chunk[str->length] = '\0';
  
  return chunk;
}


static impulse_unistring impulse_mapping_from_ascii(impulse_char *chunk)
{
  impulse_unistring str = impulse_unistring_create_space(strlen(chunk));
  impulse_uint i;
  
  for (i = 0 ; i < str->capacity ; i++)
    str->data[i] = (impulse_unichar)((chunk[i] & 0x80)? '?' : chunk[i]);
  str->data[str->capacity] = 0;
  str->length = str->capacity;
  return str;
}

static impulse_char *impulse_mapping_to_utf8(impulse_unistring str)
{
  return impulse_unistring_to_utf(str, 8, 0);
}

static impulse_unistring impulse_mapping_from_utf8(impulse_char *chunk)
{
  return impulse_unistring_create_utf(chunk, 8);
}

static impulse_char *impulse_mapping_to_utf7(impulse_unistring str)
{
  return impulse_unistring_to_utf(str, 7, 0);
}

static impulse_unistring impulse_mapping_from_utf7(impulse_char *chunk)
{
  return impulse_unistring_create_utf(chunk, 7);
}


static struct _impulse_mapping_engine ascii_engine = 
{
  1,
  (struct _impulse_mapping_engine_data) { NULL, NULL },
  (struct _impulse_mapping_engine_func) { impulse_mapping_from_ascii, impulse_mapping_to_ascii }
};
static struct _impulse_mapping_engine utf8_engine = 
{
  1,
  (struct _impulse_mapping_engine_data) { NULL, NULL },
  (struct _impulse_mapping_engine_func) { impulse_mapping_from_utf8, impulse_mapping_to_utf8 }
};
static struct _impulse_mapping_engine utf7_engine = 
{
  1,
  (struct _impulse_mapping_engine_data) { NULL, NULL },
  (struct _impulse_mapping_engine_func) { impulse_mapping_from_utf7, impulse_mapping_to_utf7 }
};

static struct _impulse_unistring ascii_encoding = 
{
  5, 5, (impulse_unichar[]) 
  {
    (impulse_unichar)'a',
    (impulse_unichar)'s',
    (impulse_unichar)'c',
    (impulse_unichar)'i',
    (impulse_unichar)'i',
    (impulse_unichar)'\0'
  }
};
static struct _impulse_unistring utf8_encoding =
{
  4, 4, (impulse_unichar[]) 
  {
    (impulse_unichar)'u',
    (impulse_unichar)'t',
    (impulse_unichar)'f',
    (impulse_unichar)'8',
    (impulse_unichar)'\0'
  }
};
static struct _impulse_unistring utf7_encoding = 
{
  4, 4, (impulse_unichar[]) 
  {
    (impulse_unichar)'u',
    (impulse_unichar)'t',
    (impulse_unichar)'f',
    (impulse_unichar)'7',
    (impulse_unichar)'\0'
  }
};


static impulse_mapping impulse_mapping_create_int(impulse_unistring encoding)
{
  impulse_char *envpath = getenv("IMPULSE_MAPPING_PATH");
  impulse_unistring path = NULL;
  impulse_module module = NULL;
  impulse_mapping map = impulse_alloc(sizeof(struct _impulse_mapping));
  void *sym;
  
  /* 
   * Setup the default mapping engine in case we can't load
   * the one requested
   */
  map->refs = 1;
  map->module = NULL;
  map->engine = NULL;
  map->encoding = NULL;
  
  if (impulse_unistring_equal(encoding, &ascii_encoding))
  {
    map->engine = &ascii_engine;
    map->encoding = &ascii_encoding;
    return map;
  }
  if (impulse_unistring_equal(encoding, &utf8_encoding))
  {
    map->engine = &utf8_engine;
    map->encoding = &utf8_encoding;
    return map;
  }
  if (impulse_unistring_equal(encoding, &utf7_encoding))
  {
    map->engine = &utf7_engine;
    map->encoding = &utf7_encoding;
    return map;
  }

  if (envpath)
  {
    path = impulse_unistring_create_utf(envpath, 8);
    impulse_unistring_append_utf(path, "/", 8);
  }
  else
  {
    path = impulse_unistring_create_utf(PREFIX_DIR, 8);
    impulse_unistring_append_utf(path, "/lib/impulse/modules/mapping/", 8);
  }
  
  impulse_unistring_append(path, encoding);
  
  if (!(module = impulse_module_create(path)))
  {
    map->engine = &ascii_engine;
    map->encoding = impulse_unistring_create(encoding);
    impulse_unistring_destroy(path);
    return map;
  }
  impulse_unistring_destroy(path);
  

  if (!(sym = impulse_module_symbol(module, "engine")))
  {
    map->engine = &ascii_engine;
    map->encoding = impulse_unistring_create(encoding);
    impulse_module_destroy(module);
    return map;
  }

  map->encoding = impulse_unistring_create(encoding);
  map->engine = (impulse_mapping_engine)sym;
  map->module = module;
  return map;
}


/*                            ======--------======                           */


impulse_mapping impulse_mapping_create(impulse_unistring encoding)
{
  impulse_mapping map = NULL;

#if USE_THREADS
  impulse_mutex_lock(&encoding_mutex);
#endif
  if (!encodings)
    encodings = impulse_hash_create();
  
  if (!(map = impulse_hash_recover(encodings, encoding)))
  {
    map = impulse_mapping_create_int(encoding);
    impulse_hash_store(encodings, encoding, map);
  }
  else
    map->refs++;
  
#if USE_THREADS
  impulse_mutex_unlock(&encoding_mutex);
#endif
  return map;
}


void impulse_mapping_destroy(impulse_mapping map)
{
#if USE_THREADS
  impulse_mutex_lock(&encoding_mutex);
#endif
  if (!encodings)
    encodings = impulse_hash_create();
  
  if (!(--(map->refs)))
  {
    impulse_hash_remove(encodings, map->encoding);
    
    if (map->module)
    {
      impulse_module_destroy(map->module);
      impulse_unistring_destroy(map->encoding);
    }
    impulse_free(map);
  }
#if USE_THREADS
  impulse_mutex_unlock(&encoding_mutex);
#endif
}


impulse_unistring impulse_mapping_to_unicode(impulse_mapping map,
					     impulse_char *chunk)
{
  if (map->engine->type)
    return map->engine->func.to(chunk);
  else
  {
    impulse_unichar *mapping = map->engine->data.to;
    impulse_unistring str = impulse_unistring_create_space(strlen(chunk));
    impulse_uint i = 0;
    
    while (*chunk)
    {
      unsigned char c = *chunk++;
      str->data[i++] = mapping[c];
    }
    str->length = str->capacity;
    str->data[str->length] = 0;
    return str;
  }
}


impulse_char *impulse_mapping_from_unicode(impulse_mapping map,
					   impulse_unistring str)
{
  if (map->engine->type)
    return map->engine->func.from(str);
  else
  {
    impulse_char **mapping = map->engine->data.from;
    impulse_char *chunk = NULL;
    impulse_uint len = 0;
    impulse_uint i;
  
    for (i = 0 ; i < str->length ; i++)
    {
      len++;
      chunk = impulse_realloc(chunk, len*sizeof(char));

      /* FIXME: No provision is made for surrogates, yet */
      if (str->data[i] >= IMPULSE_UNICHAR_SURR_HIGH_START &&
	  str->data[i] <= IMPULSE_UNICHAR_SURR_HIGH_END)
      {
	impulse_uint lowpos = i+1;
	
	chunk[len-1] = '?';

	if (lowpos < str->length-1)
	{
	  if (str->data[lowpos] >= IMPULSE_UNICHAR_SURR_LOW_START &&
	      str->data[lowpos] <= IMPULSE_UNICHAR_SURR_LOW_END)
	  {
	    i = lowpos;
	  }
	}
      }
      else
      {
	impulse_uint d1, d2;

	d1 = (str->data[i]>>8)&0xff;
	d2 = str->data[i]&0xff;

	chunk[len-1] = mapping[d1][d2];
      }
    }

    len++;
    chunk = impulse_realloc(chunk, len*sizeof(char));
    chunk[len-1] = '\0';
    return chunk;
  }
}


/* End of file */
