/*                                                                |- Impulse -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is the Impulse utility library
 * for Impulse. (Part of the ROWS distribution)
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter. Portions created by Andrew Hunter
 * are Copyright (C) 1999 Andrew Hunter.
 * All Rights Reserved
 *
 * Contributor(s):
 * Portions created by Daniel Berrange
 * are Copyright (C) 1999 Daniel Berrange.
 */

/*
 * memory.c
 * --------
 *
 * Memory handling functions
 */

#include "../../../config.h"

#include "impulse/util/memory.h"
#include "impulse/util/error.h"
#include "impulse/util/unistring.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#if HAVE_DEBUG_MCHECK
#include <mcheck.h>
#endif

#if USE_THREADS
#include "impulse/util/thread.h"

static impulse_mutex_static memory_mutex = IMPULSE_MUTEX_CREATE_STATIC;
#endif

#if HAVE_DEBUG_MCHECK
void abortfunc(enum mcheck_status stat)
{
  if (stat == MCHECK_OK ||
      stat == MCHECK_DISABLED)
    return;
  
  write(STDERR_FILENO, "*****\n", 6);
  write(STDERR_FILENO, "***** Impulse Emergency\n", 24);
  write(STDERR_FILENO, "***** Memory corruption\n", 24);
  switch (stat)
  {
   case MCHECK_FREE:
    write(STDERR_FILENO, "***** Block freed twice\n", 24);
    break;
    
   case MCHECK_HEAD:
    write(STDERR_FILENO, "***** Memory before block was clobbered\n", 40);
    break;
    
   case MCHECK_TAIL:
    write(STDERR_FILENO, "***** Memory after block was clobbered\n", 39);
    break;
    
   default:
    write(STDERR_FILENO, "***** Unknown problem\n", 22);
    break;
  }
  write(STDERR_FILENO, "*****\n", 6);
  sleep(60);
  exit(1);
}
#endif

void impulse_mcheck(void)
{
#if HAVE_DEBUG_MCHECK
  mcheck(abortfunc);
#endif

  printf("*****\n");
  printf("***** Impulse Information\n");
  printf("***** Memory checking enabled\n");
  printf("*****\n");
}

#if CHECK_MEMORY_LEAKS
static struct _impulse_mem_list  *mem = NULL;
static struct _impulse_mem_stats  stats = { 0, 0, 0, 0, 0 };
#endif

#if !CHECK_MEMORY_LEAKS
void *impulse_alloc(size_t size)
#else
void *impulse_alloc_c(size_t      size,
		      int         line_no,
		      char* const file,
		      char* const comment)
#endif
{
  impulse_memory_block *d;

  /*
   * We don't lock the memory mutex here, because nobody can use this
   * piece of memory until we've returned it!
   */
  
  d = malloc(size + sizeof(struct _impulse_memory_block));

  if (!d)
  {
    /*
     * While I would love to call our error handling functions now,
     * they would no doubt try to allocate more memory resulting in
     * bad shit happening. (infinite loops for instance)
     */

    /*
     * It'll happen anyway when another thread is switched in about
     * here... Should probably reserve an emergancy memory pool and
     * use that while we Terminate With Extreme Prejudice...
     */
    write(STDERR_FILENO, "*****\n", 6);
    write(STDERR_FILENO, "***** Impulse Emergency\n",24);
    write(STDERR_FILENO, "***** Out of memory\n", 20);
    write(STDERR_FILENO, "*****\n", 6);
    exit(1);
  }
  
#if !CHECK_MEMORY_LEAKS
  d->meta  = NULL;             // Meta data
#else
  // Store some meta-data
  d->magic = (impulse_uint) d; // Magic number
  d->meta           = malloc(sizeof(impulse_memory_meta));
  d->meta->user     = NULL;
  d->meta->allocate = malloc(sizeof(impulse_memory_allocator));

  d->meta->allocate->source_file = file;
  d->meta->allocate->line        = line_no;
  d->meta->allocate->comment     = comment;
  d->meta->allocate->size        = size;

  {
    struct _impulse_mem_list *mlist;

    mlist = malloc(sizeof(struct _impulse_mem_list));
    mlist->d    = d;
    mlist->next = mem;
    mlist->last = NULL;
    if (mem != NULL)
      mem->last = mlist;

    d->meta->allocate->item = mlist;
    mem = mlist;
  }

  stats.current += size;
  stats.total += size;
  if (stats.current > stats.maximum)
    stats.maximum = stats.current;
  stats.num_blocks++;
  if (stats.num_blocks > stats.max_blocks)
    stats.max_blocks = stats.num_blocks;
#endif

  return (void*) d + sizeof(struct _impulse_memory_block);;
}

/*
 * ---*** Danger, Will Robinson! ***---
 *
 * Though impulse_free is now thread-safe, assuming you've kept track
 * of which threads are using which piece of memory, Bad Things can
 * happen if you realloc without telling the other threads
 *
 * Maybe the realloc should fail if you try it on a piece of memory
 * that is in use by other threads, or maybe the old piece of memory
 * should continue to exist (fairly yucky, that)
 */
#if CHECK_MEMORY_LEAKS
void *impulse_realloc_c(void       *old,
			size_t      size,
			int         line_no,
			char *const file)
#else
void *impulse_realloc(void *old, size_t size)
#endif
{
  impulse_memory_block *d;

  impulse_mutex_lock(&memory_mutex);
  
  if (old != NULL)
    {
      d = old - sizeof(struct _impulse_memory_block);

#if CHECK_MEMORY_LEAKS
      if (d->magic != (impulse_uint) d)
	{
	  write(STDERR_FILENO, "*****\n", 6);
	  write(STDERR_FILENO, "***** Impulse Warning\n",22);
	  write(STDERR_FILENO, "***** Block passed to impulse_realloc() wasn't allocated by Impulse\n", 68);
	  write(STDERR_FILENO, "*****\n", 6);
	}
#endif
	  
      d = realloc(d, size + sizeof(struct _impulse_memory_block));
      
      if (!d)
	{
	  /*
	   * cf. impulse_alloc
	   */
	  write(STDERR_FILENO, "*****\n", 6);
	  write(STDERR_FILENO, "***** Impulse Emergency\n",24);
	  write(STDERR_FILENO, "***** Out of memory\n", 20);
	  write(STDERR_FILENO, "*****\n", 6);
	  exit(1);
	}

#if CHECK_MEMORY_LEAKS
      d->magic = (impulse_uint) d; // Magic number

      d->meta->allocate->item->d = d;

      if (size > d->meta->allocate->size)
	stats.total += (size - d->meta->allocate->size);
      else
	stats.total_freed += (d->meta->allocate->size - size);
      stats.current -= d->meta->allocate->size;
      stats.current += size;
      if (stats.current > stats.maximum)
	stats.maximum = stats.current;

      d->meta->allocate->size = size;
#endif
      
      impulse_mutex_unlock(&memory_mutex);
      
      return (void*) d + sizeof(struct _impulse_memory_block);
    }
  else
    {
      impulse_mutex_unlock(&memory_mutex);

#if CHECK_MEMORY_LEAKS
      return impulse_alloc_c(size, line_no, file, "Allocated by impulse_realloc()");
#else
      return impulse_alloc(size);
#endif
    }
}

void impulse_free(void *old)
{
  impulse_memory_block *check;
  
  impulse_mutex_lock(&memory_mutex);

  if (old)
    {
      check = old - sizeof(impulse_memory_block);

#if CHECK_MEMORY_LEAKS
      if (check->magic != (impulse_uint) check)
	{
	  write(STDERR_FILENO, "*****\n", 6);
	  write(STDERR_FILENO, "***** Impulse Warning\n",22);
	  write(STDERR_FILENO, "***** Block passed to impulse_free() wasn't allocated by Impulse\n", 65);
	  write(STDERR_FILENO, "*****\n", 6);

	  free(old);
	}
      else
#endif
	{
	  int dofree;

	  dofree = 1;
	  
	  if (check->meta != NULL)
	    {
	      dofree = (--check->meta->count) < 0;
	    }

	  if (dofree)
	    {
#if CHECK_MEMORY_LEAKS
	      check->magic = 0xdeadf00d;
	      stats.total_freed += check->meta->allocate->size;
	      stats.current     -= check->meta->allocate->size;
	      stats.num_blocks--;
	      
	      if (check->meta->allocate->item->next != NULL)
		check->meta->allocate->item->next->last =
		  check->meta->allocate->item->last;
	      if (check->meta->allocate->item->last != NULL)
		check->meta->allocate->item->last->next =
		  check->meta->allocate->item->next;
	      else
		mem = check->meta->allocate->item->next;

	      free(check->meta->allocate->item);
	      free(check->meta->allocate);
#endif
	      if (check->meta)
		free(check->meta);
	      free(check);
	    }
	}
    }
  else
    {
      write(STDERR_FILENO, "*****\n", 6);
      write(STDERR_FILENO, "***** Impulse Warning\n",22);
      write(STDERR_FILENO, "***** Tried to free NULL\n", 25);
      write(STDERR_FILENO, "*****\n", 6);      
    }

  impulse_mutex_unlock(&memory_mutex);
}

#if CHECK_MEMORY_LEAKS
struct _impulse_mem_stats  impulse_memory_get_stats(void)
{
  return stats;
}

struct _impulse_mem_list  *impulse_memory_get_allocated(void)
{
  return mem;
}
#endif

/* End of file */
