#!/usr/bin/perl

#                                                                 |- Impulse -|
# The contents of this file are subject to the Mozilla
# Public License Version 1.0 (the "License"); you may
# not use this file except in compliance with the
# License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/ 
#
# Software distributed under the License is distributed
# on an "AS IS" basis, WITHOUT WARRANTY OF
# ANY KIND, either express or implied. See the
# License for the specific language governing rights
# and limitations under the License. 
#
# The Original Code is the ROWS widget library
# for Impulse.
#
# The Initial Developer of the Original Code is
# Andrew Hunter. Portions created by Andrew Hunter
# are Copyright (C) 1999 Andrew Hunter.
# All Rights Reserved
#
# Contributor(s):
#  Portions created by Daniel Berrange
# are Copyright (C) 1999 Daniel Berrange.
# All Rights Reserved# No-one (yet :-)
#

#
# Perl script for creating C files containing 8-bit to Unicode mapping
# descriptions.
#
# Syntax: mapping2c.pl <map file name> <c file name>
#

#
# Uses the file format set down by the files supplied by Unicode, Inc. on the
# Unicode Standard 2.0 CDROM. That is:
#
# Lines beginning with a '#' are comments
# Empty lines (containing only whitespace) are ignored
# Other lines contain 3 fields, seperated by whitespace - 8-bit value &
#   Unicode value (in hexidecimal, prefixed by 0x)
#

use IO::Handle;
use File::Basename;
use POSIX;

$infile     = $ARGV[0];
$outcfile   = $ARGV[1];

$offset1 = rindex $outcfile, "/";
$offset2 = rindex $outcfile, ".";
$mapvarname = substr $outcfile, $offset1+1, $offset2 - $offset1 - 1;
$mapname    = $mapvarname;

print STDERR "mapping2c.pl v0.2 written by Andrew Hunter (updated by Daniel Berrange)\n";

# Read the data

open INFILE, "<$infile";

print STDERR "Reading mapping data from $infile.";
flush STDERR;

$chars = 0;

while (<INFILE>)
  {
    if (/^\s*$/)
      {
	# Whitespace. Continue
      }
    elsif (/^\s*\#(.*)$/)
      {
	# Comment. Continue
      }
    elsif (/^\s*0x([0-9A-Fa-f]+)\s+0x([0-9A-Fa-f]+)\s+(.*)$/)
      {
	# Character mapping
	$chars++;
	$eight = hex($1);
	$uni   = hex($2);

	if ($eight < 0   || $uni < 0 ||
	    $eight > 255 || $uni > 65535)
	  {
	    die "\nSyntax error at line $.\n";
	  }
	
	$eightuni[$eight] = $uni;
	$unieight{$uni}   = $eight;

	$occupied[floor($uni/256)] = 1;	   

	if (($chars%16) == 0)
	  {
	    print STDERR ".";
	    flush STDERR;
	  }
      }
    else
      {
	die "\nSyntax error at line $.\n";
      }
  }

print STDERR "\n";
print STDERR "Read $chars characters\n";

close INFILE;

for ($char=0; $char<256; $char++)
  {
    if (!defined($eightuni[$char]))
      {
	# Control character/direct mapping
	$eightuni[$char] = $char;
	$unieight{$char} = $char;
	$occupied[0] = 1;
      }
  }

# Output the C file

print STDERR "Generating C file...\n";

open CFILE, ">$outcfile";
print CFILE "/*\n * C file automatically generated by mapping2c.pl\n * Mapping describes the $mapname mapping\n * The contents of this file may be based on data under differing license\n * agreements from the rest of libImpLib. See the original mapping file for \n * any relevant details\n *\n * mapping2c.pl was written by Andrew Hunter\n */\n\n";
print CFILE "#include <stdio.h>\n";
print CFILE "#include <impulse/util/mapping_int.h>\n\n\n";

# Unicode to 8-bit structures

print CFILE "static impulse_char U8_empty[256] =\n";
print CFILE "{\n";
for ($x=0; $x<256; $x++)
  {
    print CFILE "    '?',\n";
  }
print CFILE "};\n";

for ($section=0; $section<256; $section++)
  {
    if ($occupied[$section])
      {
	printf CFILE "static impulse_char U8_%04x\[256] =\n", $section*256;
	print CFILE "{\n";
	$start = $section*256;
	$end   = $start+256;
	for ($char=$start; $char<$end; $char++)
	  {
	    if ($char < 32)
	      {
		# Control code
		$unieight{$char} = $char;
	      }

            if (defined $unieight{$char}) {
              printf CFILE "    0x%02x,\n", $unieight{$char};
	    } else {
	      print CFILE "    '?',\n";
	    }
	  }
	print CFILE "};\n";
      }
  }

# 8-bit to unicode data
print CFILE "static impulse_unichar to_unicode[256] = \n{\n";
for ($char=0; $char<256; $char++)
  {
    printf CFILE "        0x%04x,\n", $eightuni[$char];
  }
print CFILE "};\n";

# Unicode to 8-bit data
print CFILE "static impulse_char *from_unicode[256] = \n{\n";
for ($section=0;$section<256;$section++)
  {
    if ($occupied[$section] == 1)
      {
	printf CFILE "U8_%04x", $section*256;
      }
    else
      {
	printf CFILE "U8_empty";
      }

    if ($section<255)
      {
	print CFILE ", ";
	if ((($section+1)%8) == 0)
	  {
	    print CFILE "\n      ";
	  }
      }
    else
      {
	print CFILE "\n";
      }
  }

print CFILE "};\n";

print CFILE "struct _impulse_mapping_engine $mapvarname" . "_engine = {\n";
print CFILE " 0, {to_unicode, from_unicode}, {NULL,NULL}\n";
print CFILE "};\n";

print CFILE "/* End of file */\n\n";

close CFILE;

# Finish up

print STDERR "That seems to be it\n";
