/*                                                                  |- CFSFC -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is Copyright (C) 1998 Daniel Berrange.
 * All Rights Reserved. 
 *
 * The Initial Developer of the Original Code is
 * Daniel Berrange.
 *
 * Contributor(s):
 * No-one yet
 */

/*
 * impulse_bug.c
 * -------------
 *
 * Provide complex user interaction when program crashes.
 */


#include <impulse/util/main.h>
#include <impulse/util/text.h>
#include <impulse/util/memory.h>
#include <impulse/util/error.h>
#include "message.h"

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/wait.h>

void do_args(void);
void do_query(void);
void do_action(void);
void do_debug(void);
void do_trace(void);
void do_report(void);

static impulse_uint level;
static impulse_unistring app;
static impulse_unistring msg;
static impulse_unistring command;
static impulse_uint pid;

static impulse_uint in;
static impulse_uint out;


impulse_int main(impulse_int argc, impulse_char *argv[])
{
  if (getenv("IMPULSE_ERROR_HANDLER"))
    exit(-1);
  putenv("IMPULSE_ERROR_HANDLER=1");

  impulse_error_handle(IMPULSE_ERROR_EMERGENCY, impulse_error_print);
  impulse_error_handle(IMPULSE_ERROR_CRITICAL, impulse_error_print);
  impulse_error_handle(IMPULSE_ERROR_WARNING, impulse_error_print);
  impulse_error_handle(IMPULSE_ERROR_INFO, impulse_error_print);
  impulse_init(argc, argv, "impulse_bug");
  
  impulse_bug_text = impulse_text_create(impulse_bug_text_base,
					 impulse_app_locale());

  do_args();
  do_query();
  do_action();
  
  impulse_text_destroy(impulse_bug_text);
  impulse_cleanup();
  return 0;
}

  
void do_args(void)
{
  impulse_unistring *uin;
  impulse_unistring *uout;
  impulse_char *cin;
  impulse_char *cout;

  impulse_command_option(_(TEXT_OPTION_IN), 0,
			 _(TEXT_HELP_IN),
			 IMPULSE_APP_ARG_SINGLE);
  impulse_command_option(_(TEXT_OPTION_OUT), 0, 
			 _(TEXT_HELP_OUT),
			 IMPULSE_APP_ARG_SINGLE);
  
  impulse_command_parse();
  
  if (impulse_command_argument(_(TEXT_OPTION_IN),
			       &uin) != 1)
    exit(1);
  if (impulse_command_argument(_(TEXT_OPTION_OUT),
			       &uout) != 1)
    exit(1);
  
  cin = impulse_unistring_to_utf(uin[0], 8, 0);
  cout = impulse_unistring_to_utf(uout[0], 8, 0);
  in = atoi(cin);
  out = atoi(cout);
  impulse_free(cin);
  impulse_free(cout);
}


void do_query(void)
{
  impulse_int len;
  
  pid = getppid();
  
  /* Get error level */
  if (read(in, &level, sizeof(level)) <= 0)
    exit(0);
  
  /* Get app name */
  if (read(in, &len, sizeof(len)) <= 0)
    exit(0);
  app = impulse_unistring_create_space(len);
  if (read(in, app->data, sizeof(impulse_unichar)*(len+1)) <= 0)
    exit(0);
  app->length = app->capacity;

  /* Get command name */
  if (read(in, &len, sizeof(len)) <= 0)
    exit(0);
  command = impulse_unistring_create_space(len);
  if (read(in, command->data, sizeof(impulse_unichar)*(len+1)) <= 0)
    exit(0);
  command->length = command->capacity;

  /* Get error message */
  if (read(in, &len, sizeof(len)) <= 0)
    exit(0);
  msg = impulse_unistring_create_space(len);
  if (read(in, msg->data, sizeof(impulse_unichar)*(len+1)) <= 0)
    exit(0);
  msg->length = msg->capacity;
}


void do_action(void)
{
  impulse_char *cchoice;
  impulse_char c;
  
  cchoice = impulse_unistring_to_8bit(_(TEXT_ACTION_CHOICE),
				      impulse_app_charset());

  impulse_error_print(level, msg);

  for (;;)
  {
    printf(cchoice);
    printf("> ");
    fflush(stdout);
    
    scanf("%c", &c);
    
    if (feof(stdin)) 
    {
      exit(0);  
    }
    
    switch (c)
    {
     case 'E':
     case 'e':
     case 'x':
     case 'X':
     case 'q':
     case 'Q':
      exit(0);
      
     case 'D':
     case 'd':
      do_debug();
      exit(0);
      
     case 'T':
     case 't':
      do_trace();
      exit(0);
      
     case 'B':
     case 'b':
      do_report();
      exit(0);
      
     default:
      continue;
    }
  }
  impulse_free(cchoice);
}
  

void do_debug(void)
{
  int child;
  char cpid[50];
  char *prog = impulse_unistring_to_8bit(command, impulse_app_charset());
  char *args[5];

  printf("*****\n");
  printf("***** Thanks to some wierd interactions between gdb and pthreads\n");
  printf("***** you *MUST* issue the kill command before exiting gdb. Otherwise\n");
  printf("***** the debugger will hang indefinately\n");
  printf("*****\n");
  snprintf(cpid, 49, "%d", pid);
  
  args[0] = "gdb";
  args[1] = "-q";
  args[2] = prog;
  args[3] = cpid;
  args[4] = NULL;
  
  child = fork();
  if (!child)
  {
    execvp(args[0], args);
    /* Should never reach here */
    _exit(1);
  }
  else if (child < 0)
    return; /* Do something better*/

  waitpid(child, NULL, 0);
  impulse_free(prog);
}


void do_trace(void)
{
  int child;
  int new_stdin[2];
  int new_stdout[2];
  char cpid[50];
  char *prog = impulse_unistring_to_8bit(command, impulse_app_charset());
  char *args[5];
  
  snprintf(cpid, 49, "%d", pid);

  args[0] = "gdb";
  args[1] = "-n";
  args[2] = prog;
  args[3] = cpid;
  args[4] = NULL;

  if (pipe(new_stdin) < 0)
    return;
  
  if (pipe(new_stdout) < 0)
  {
    close(new_stdin[0]);
    close(new_stdin[1]);
  }

  child = fork();
  if (!child)
  {
    dup2(new_stdin[0], STDIN_FILENO);
    dup2(new_stdout[1], STDOUT_FILENO);
    dup2(new_stdout[1], STDERR_FILENO);
    
    execvp(args[0], args);
    /* Should never reach here */
    _exit(1);
  }
  else if (child < 0)
    return; /* Do something better*/

  write(new_stdin[1], "backtrace\n", 10);
  write(new_stdin[1], "kill\n", 5);
  write(new_stdin[1], "quit\n", 5);
  write(new_stdin[1], "y\n", 2);

  for (;;)
  {
    char c;
    if (read(new_stdout[0], &c, 1) != 1)
      break;
    write(STDOUT_FILENO, &c, 1);
  }
  waitpid(child, NULL, 0);

  close(new_stdout[0]);
  close(new_stdout[1]);
  close(new_stdin[0]);
  close(new_stdin[1]);
  impulse_free(prog);
}


void do_report(void)
{
  impulse_char *report = impulse_unistring_to_8bit(_(TEXT_REPORT),
						   impulse_app_charset());
  printf(report);
  impulse_free(report);
}

  
/* End of file */
